from __future__ import annotations

import logging
import math
from dataclasses import dataclass

import numpy as np
from scipy import interpolate, stats

from cctbx import miller
from cctbx.french_wilson import (
    ac_zf,
    ac_zf_sd,
    ac_zj,
    ac_zj_sd,
    c_zf,
    c_zf_sd,
    c_zj,
    c_zj_sd,
)
from dxtbx import flumpy

logger = logging.getLogger(__name__)


def french_wilson(
    merged_intensities: miller.array,
    max_bins: int = 60,
    sigma_iobs_rejection_criterion: float = -4.0,
) -> miller.array:
    assert merged_intensities.is_xray_intensity_array()
    assert merged_intensities.sigmas() is not None
    assert merged_intensities.is_unique_set_under_symmetry()

    intensities = flumpy.to_numpy(merged_intensities.data())
    sigmas = flumpy.to_numpy(merged_intensities.sigmas())
    is_centric = flumpy.to_numpy(merged_intensities.centric_flags().data())
    is_acentric = ~is_centric

    rejected = (sigmas <= 0) & ((intensities <= 0) | (sigmas < 0))
    logger.debug(
        f"Rejecting {np.count_nonzero(rejected)} reflections with non-positive sigmas"
    )

    # Compute expected intensities
    d_star_cubed = flumpy.to_numpy(merged_intensities.d_star_cubed().data())
    bins = determine_binning_counting_sorted(d_star_cubed, max_bins=max_bins)
    expected_intensities = compute_expected_intensities(
        d_star_cubed,
        intensities,
        bins,
        m_estimator=standardized_median,
    )
    expected_intensity_is_zero = expected_intensities == 0
    logger.debug(
        f"Rejecting {np.count_nonzero(expected_intensity_is_zero)} "
        "reflections with expected intensity equal to zero"
    )
    rejected |= expected_intensity_is_zero
    valid = ~rejected

    J = np.zeros_like(intensities)
    sigJ = np.zeros_like(intensities)
    F = np.zeros_like(intensities)
    sigF = np.zeros_like(intensities)

    # Compute acentric moments
    if np.count_nonzero(is_acentric):
        posterior_moments_acentric = compute_posterior_moments_acentric(
            intensities[is_acentric],
            sigmas[is_acentric],
            expected_intensities[is_acentric],
        )
        J[is_acentric] = posterior_moments_acentric.J
        sigJ[is_acentric] = posterior_moments_acentric.sigJ
        F[is_acentric] = posterior_moments_acentric.F
        sigF[is_acentric] = posterior_moments_acentric.sigF
        valid[is_acentric] &= posterior_moments_acentric.valid

    # Compute centric moments
    if np.count_nonzero(is_centric):
        posterior_moments_centric = compute_posterior_moments_centric(
            intensities[is_centric],
            sigmas[is_centric],
            expected_intensities[is_centric],
        )
        J[is_centric] = posterior_moments_centric.J
        sigJ[is_centric] = posterior_moments_centric.sigJ
        F[is_centric] = posterior_moments_centric.F
        sigF[is_centric] = posterior_moments_centric.sigF
        valid[is_centric] &= posterior_moments_centric.valid

    logger.debug(f"Rejected {np.count_nonzero(~valid)} reflections")
    return (
        merged_intensities.customized_copy(
            data=flumpy.from_numpy(F),
            sigmas=flumpy.from_numpy(sigF),
            anomalous_flag=merged_intensities.anomalous_flag(),
        )
        .set_observation_type_xray_amplitude()
        .select(flumpy.from_numpy(valid))
    )


def standardized_median(a: np.ndarray) -> float:
    return np.median(a) / math.log(2)


def determine_binning_counting_sorted(x: np.ndarray, max_bins=60, min_bin_size=20):
    n_points = len(x)
    points_per_bin = max(len(x) // max_bins, min_bin_size)
    n_bins = len(x) // points_per_bin
    logger.debug(f"{n_bins=}, {points_per_bin=}")
    return np.interp(
        np.linspace(0, n_points, n_bins + 1), np.arange(n_points), np.sort(x)
    )


def compute_expected_intensities(
    x: np.ndarray,
    intensities: np.ndarray,
    bins: int | list[int | float],
    m_estimator="mean",
) -> np.ndarray:
    # Compute the mean intensities binned according to x
    bin_means, bin_edges, binnumber = stats.binned_statistic(
        x, intensities, statistic=m_estimator, bins=bins
    )
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2

    # Interpolate expected intensities
    # f = interpolate.interp1d(bin_centers, bin_means, fill_value="extrapolate")
    f = interpolate.interp1d(
        bin_centers,
        bin_means,
        bounds_error=False,
        fill_value=(bin_means[0], bin_means[-1]),
    )
    # f = interpolate.CubicSpline(bin_centers, bin_means, bc_type="clamped")
    expected_intensities = f(x)
    return expected_intensities


@dataclass
class PosteriorMoments:
    J: np.ndarray
    sigJ: np.ndarray
    F: np.ndarray
    sigF: np.ndarray
    valid: np.ndarray


def compute_posterior_moments_acentric(
    intensities: np.ndarray,
    sigmas: np.ndarray,
    expected_intensities: np.ndarray,
    h_min: int = -4,
) -> PosteriorMoments:

    h = (intensities / sigmas) - np.abs(sigmas / expected_intensities)
    logger.debug(f"h range: {h.min():.4f} - {h.max():.4f}")
    i_sig_min = h_min + 0.3
    valid = (intensities / sigmas >= i_sig_min) & (h >= h_min)

    J = np.empty_like(h)
    sigJ = np.empty_like(h)
    F = np.empty_like(h)
    sigF = np.empty_like(h)

    # Set up interpolation functions as a function of h
    # Tables give values from h = -4.0 to h = 10.0 in steps of 0.1
    x = (np.array(range(len(ac_zj)))) / 10 - 4
    J_h = interpolate.interp1d(x, ac_zj)
    sigJ_h = interpolate.interp1d(x, ac_zj_sd)
    F_h = interpolate.interp1d(x, ac_zf)
    sigF_h = interpolate.interp1d(x, ac_zf_sd)

    # Interpolation only valid for -4 < h < 3
    sel = valid & (h < 3)
    J[sel] = J_h(h[sel]) * sigmas[sel]
    sigJ[sel] = sigJ_h(h[sel]) * sigmas[sel]
    F[sel] = F_h(h[sel]) * np.sqrt(sigmas)[sel]
    sigF[sel] = sigF_h(h[sel]) * np.sqrt(sigmas)[sel]

    # For h >= 3 approximate as a normal distribution
    sel = valid & (h >= 3)
    J[sel] = h[sel] * sigmas[sel]
    sigJ[sel] = sigmas[sel]
    F[sel] = np.sqrt(J[sel])
    sigF[sel] = 0.5 * sigmas[sel] / F[sel]

    return PosteriorMoments(
        J=J,
        sigJ=sigJ,
        F=F,
        sigF=sigF,
        valid=valid,
    )


def compute_posterior_moments_centric(
    intensities: np.ndarray,
    sigmas: np.ndarray,
    expected_intensities: np.ndarray,
    h_min: int = -4,
) -> PosteriorMoments:

    h = (intensities / sigmas) - np.abs(sigmas / (2 * expected_intensities))
    logger.debug(f"h range: {h.min():.4f} - {h.max():.4f}")
    i_sig_min = h_min + 0.3
    valid = (intensities / sigmas >= i_sig_min) & (h >= h_min)

    J = np.empty_like(h)
    sigJ = np.empty_like(h)
    F = np.empty_like(h)
    sigF = np.empty_like(h)

    # Set up interpolation functions as a function of h
    # Tables give values from h = -4.0 to h = 10.0 in steps of 0.1
    x = (np.array(range(len(c_zj)))) / 10 - 4
    J_h = interpolate.interp1d(x, c_zj)
    sigJ_h = interpolate.interp1d(x, c_zj_sd)
    F_h = interpolate.interp1d(x, c_zf)
    sigF_h = interpolate.interp1d(x, c_zf_sd)

    # Obtain posterior moments via interpolation for -4 < h < 4
    sel = valid & (h < 4)
    J[sel] = J_h(h[sel]) * sigmas[sel]
    sigJ[sel] = sigJ_h(h[sel]) * sigmas[sel]
    F[sel] = F_h(h[sel]) * np.sqrt(sigmas)[sel]
    sigF[sel] = sigF_h(h[sel]) * np.sqrt(sigmas)[sel]

    # adapted from French-Wilson w/ added x^6 term in the expansion
    sel = valid & (h >= 4)
    h_2 = 1 / np.square(h[sel])
    h_4 = np.square(h_2)
    h_6 = h_2 * h_4
    posterior_F = np.sqrt(h[sel]) * (
        1 - (3 / 8) * h_2 - (87 / 128) * h_4 - (2889 / 1024) * h_6
    )
    posterior_sigF = np.sqrt(
        h[sel] * ((1 / 4) * h_2 + (15 / 32) * h_4 + (273 / 128) * h_6)
    )
    J[sel] = h[sel] * sigmas[sel] * (1 - (1 / 2) * h_2 - (3 / 4) * h_4 - 3 * h_6)
    sigJ[sel] = 2 * sigmas[sel] * posterior_F * posterior_sigF
    F[sel] = posterior_F * np.sqrt(sigmas[sel])
    sigF[sel] = posterior_sigF * np.sqrt(sigmas[sel])

    return PosteriorMoments(
        J=J,
        sigJ=sigJ,
        F=F,
        sigF=sigF,
        valid=valid,
    )



"""Analysis of centroid residuals for determining suitable refinement and
outlier rejection parameters automatically"""


from __future__ import annotations

import math

from scitbx.math.periodogram import Periodogram

from dials.array_family import flex

RAD2DEG = 180.0 / math.pi


class CentroidAnalyser:
    def __init__(self, reflections, av_callback=flex.mean, debug=False):

        # flags to indicate at what level the analysis has been performed
        self._average_residuals = False
        self._spectral_analysis = False

        self._av_callback = av_callback

        # Remove invalid reflections
        reflections = reflections.select(~(reflections["miller_index"] == (0, 0, 0)))
        x, y, z = reflections["xyzcal.mm"].parts()
        sel = (x == 0) & (y == 0)
        reflections = reflections.select(~sel)
        self._nexp = flex.max(reflections["id"]) + 1

        # Ensure required keys are present
        if not all(k in reflections for k in ["x_resid", "y_resid", "phi_resid"]):
            x_obs, y_obs, phi_obs = reflections["xyzobs.mm.value"].parts()
            x_cal, y_cal, phi_cal = reflections["xyzcal.mm"].parts()

            # do not wrap around multiples of 2*pi; keep the full rotation
            # from zero to differentiate repeat observations.

            TWO_PI = 2.0 * math.pi
            resid = phi_cal - (flex.fmod_positive(phi_obs, TWO_PI))
            # ensure this is the smaller of two possibilities
            resid = flex.fmod_positive((resid + math.pi), TWO_PI) - math.pi
            phi_cal = phi_obs + resid
            reflections["x_resid"] = x_cal - x_obs
            reflections["y_resid"] = y_cal - y_obs
            reflections["phi_resid"] = phi_cal - phi_obs

        # create empty results list
        self._results = []

        # first, just determine a suitable block size for analysis
        for iexp in range(self._nexp):
            ref_this_exp = reflections.select(reflections["id"] == iexp)
            if len(ref_this_exp) == 0:
                # can't do anything, just keep an empty dictionary
                self._results.append({})
                continue
            phi_obs_deg = ref_this_exp["xyzobs.mm.value"].parts()[2] * RAD2DEG
            phi_range = flex.min(phi_obs_deg), flex.max(phi_obs_deg)
            phi_width = phi_range[1] - phi_range[0]
            ideal_block_size = 1.0
            old_nblocks = 0
            while True:
                nblocks = int(phi_width // ideal_block_size)
                if nblocks == old_nblocks:
                    nblocks -= 1
                nblocks = max(nblocks, 1)
                block_size = phi_width / nblocks
                nr = flex.int()
                for i in range(nblocks - 1):
                    blk_start = phi_range[0] + i * block_size
                    blk_end = blk_start + block_size
                    sel = (phi_obs_deg >= blk_start) & (phi_obs_deg < blk_end)
                    nref_in_block = sel.count(True)
                    nr.append(nref_in_block)
                # include max phi in the final block
                blk_start = phi_range[0] + (nblocks - 1) * block_size
                blk_end = phi_range[1]
                sel = (phi_obs_deg >= blk_start) & (phi_obs_deg <= blk_end)
                nref_in_block = sel.count(True)
                nr.append(nref_in_block)
                # Break if there are enough reflections, otherwise increase block size,
                # unless only one block remains
                if nblocks == 1:
                    break
                min_nr = flex.min(nr)
                if min_nr >= 50:
                    break
                if min_nr < 5:
                    fac = 2
                else:
                    fac = 50 / min_nr
                ideal_block_size *= fac
                old_nblocks = nblocks

            # collect the basic data for this experiment
            self._results.append(
                {
                    "block_size": block_size,
                    "nref_per_block": nr,
                    "nblocks": nblocks,
                    "phi_range": phi_range,
                }
            )

        # keep reflections for analysis
        self._reflections = reflections

        # for debugging, write out reflections used
        if debug:
            self._reflections.as_file("centroid_analysis.refl")

    def __call__(
        self, calc_average_residuals=True, calc_periodograms=True, spans=(4, 4)
    ):
        """Perform analysis and return the results as a list of dictionaries (one
        for each experiment)"""

        # if not doing further analysis, return the basic data
        if not calc_average_residuals and not calc_periodograms:
            return self._results

        # if we don't have average residuals already, calculate them
        if not self._average_residuals:
            for iexp in range(self._nexp):
                results_this_exp = self._results[iexp]
                block_size = results_this_exp.get("block_size")
                if block_size is None:
                    continue
                phi_range = results_this_exp["phi_range"]
                nblocks = results_this_exp["nblocks"]
                ref_this_exp = self._reflections.select(self._reflections["id"] == iexp)
                x_resid = ref_this_exp["x_resid"]
                y_resid = ref_this_exp["y_resid"]
                phi_resid = ref_this_exp["phi_resid"]
                phi_obs_deg = ref_this_exp["xyzobs.mm.value"].parts()[2] * RAD2DEG
                xr_per_blk = flex.double()
                yr_per_blk = flex.double()
                pr_per_blk = flex.double()
                for i in range(nblocks - 1):
                    blk_start = phi_range[0] + i * block_size
                    blk_end = blk_start + block_size
                    sel = (phi_obs_deg >= blk_start) & (phi_obs_deg < blk_end)
                    xr_per_blk.append(self._av_callback(x_resid.select(sel)))
                    yr_per_blk.append(self._av_callback(y_resid.select(sel)))
                    pr_per_blk.append(self._av_callback(phi_resid.select(sel)))
                # include max phi in the final block
                blk_start = phi_range[0] + (nblocks - 1) * block_size
                blk_end = phi_range[1]
                sel = (phi_obs_deg >= blk_start) & (phi_obs_deg <= blk_end)
                xr_per_blk.append(self._av_callback(x_resid.select(sel)))
                yr_per_blk.append(self._av_callback(y_resid.select(sel)))
                pr_per_blk.append(self._av_callback(phi_resid.select(sel)))
                # the first and last block of average residuals (especially those in
                # phi) are usually bad because rocking curves are truncated at the
                # edges of the scan. When we have enough blocks and they are narrow,
                # just replace the extreme values with their neighbours
                if nblocks > 2 and block_size < 3.0:
                    xr_per_blk[0] = xr_per_blk[1]
                    xr_per_blk[-1] = xr_per_blk[-2]
                    yr_per_blk[0] = yr_per_blk[1]
                    yr_per_blk[-1] = yr_per_blk[-2]
                    pr_per_blk[0] = pr_per_blk[1]
                    pr_per_blk[-1] = pr_per_blk[-2]

                results_this_exp["av_x_resid_per_block"] = xr_per_blk
                results_this_exp["av_y_resid_per_block"] = yr_per_blk
                results_this_exp["av_phi_resid_per_block"] = pr_per_blk
            self._average_residuals = True

        # Perform power spectrum analysis on the residuals, converted to microns
        # and mrad to avoid tiny numbers
        if calc_periodograms:
            if self._spectral_analysis:
                return self._results

            for exp_data in self._results:
                exp_data["x_periodogram"] = None
                exp_data["y_periodogram"] = None
                exp_data["phi_periodogram"] = None
                if exp_data["nblocks"] < 5:
                    continue

                for pname, data in zip(
                    ["x_periodogram", "y_periodogram", "phi_periodogram"],
                    [
                        exp_data["av_x_resid_per_block"],
                        exp_data["av_y_resid_per_block"],
                        exp_data["av_phi_resid_per_block"],
                    ],
                ):
                    if (flex.max(data) - flex.min(data)) > 1.0e-8:
                        exp_data[pname] = Periodogram(1000.0 * data, spans=spans)
            self._spectral_analysis = True

            # extract further information from the power spectrum
            for exp_data in self._results:
                exp_data["x_interval"] = self._analyse_periodogram(
                    exp_data["x_periodogram"]
                )
                exp_data["y_interval"] = self._analyse_periodogram(
                    exp_data["y_periodogram"]
                )
                exp_data["phi_interval"] = self._analyse_periodogram(
                    exp_data["phi_periodogram"]
                )

        return self._results

    def _analyse_periodogram(self, pgram):
        """Use the periodogram pgram to suggest a suitable interval width for
        scan-varying refinement to account for the major variation in residuals"""

        if pgram is None:
            return None

        # determine a baseline from the high frequency noise
        bl = flex.median(pgram.spec.select(pgram.freq > 0.25))

        # look for peaks greater than 5 times this baseline. We expect one at
        # low frequency
        cutoff = 5 * bl
        peaks = pgram.spec > cutoff

        # find where this peak falls off below the cutoff and return the cycle
        # period at half that frequency (this is a heuristic that often seems to
        # give sensible results)
        pk_start = flex.first_index(peaks, True)
        if pk_start is None:
            return None
        peaks = peaks[pk_start:]
        idx = pk_start + flex.first_index(peaks, False) - 1
        if idx is not None:
            f1 = pgram.freq[idx]
            s1 = pgram.spec[idx]
            try:
                f2 = pgram.freq[idx + 1]
                s2 = pgram.spec[idx + 1]
                ds = cutoff - s1
                df = (f2 - f1) * ds / (s2 - s1)
                freq = f1 + df
            except IndexError:
                freq = f1
            period = 2.0 * 1.0 / freq
        else:
            period = None
        return period



from __future__ import annotations

import numpy as np
import pytest

from cctbx import sgtbx
from scitbx.array_family import flex

from dials.algorithms.symmetry.cosym import engine, target
from dials.algorithms.symmetry.cosym._generate_test_data import generate_test_data


@pytest.mark.parametrize("space_group", ["P2", "P3", "P6", "R3:h", "I23"])
def test_cosym_target(space_group):
    datasets, expected_reindexing_ops = generate_test_data(
        space_group=sgtbx.space_group_info(symbol=space_group).group(), sample_size=50
    )

    intensities = datasets[0]
    dataset_ids = np.zeros(intensities.size() * len(datasets))
    for i, d in enumerate(datasets[1:]):
        i += 1
        intensities = intensities.concatenate(d, assert_is_similar_symmetry=False)
        dataset_ids[i * d.size() : (i + 1) * d.size()] = np.full(d.size(), i, dtype=int)

    for weights in [None, "count", "standard_error"]:
        print(weights)
        t = target.Target(intensities, dataset_ids, weights=weights)
        m = len(t.sym_ops)
        n = len(datasets)
        assert t.dim == m
        assert t.rij_matrix.shape == (n * m, n * m)
        # x = np.random.rand(n * m * t.dim)
        x = flex.random_double(n * m * t.dim).as_numpy_array()
        f0 = t.compute_functional(x)
        g = t.compute_gradients(x)
        g_fd = t.compute_gradients_fd(x)
        np.testing.assert_allclose(g, g_fd, rtol=2e-3)
        c = t.curvatures(x)
        c_fd = t.curvatures_fd(x, eps=1e-3)
        assert list(c) == pytest.approx(c_fd, rel=0.8e-1)

        if weights == "count":
            # Absolute upper limit on weights
            assert t.wij_matrix.max() <= datasets[0].size()

        minimizer = engine.lbfgs_with_curvs(target=t, coords=x)
        # check functional has decreased and gradients are approximately zero
        f = t.compute_functional(minimizer.coords)
        g = t.compute_gradients(minimizer.coords)
        g_fd = t.compute_gradients_fd(minimizer.coords)
        assert f < f0
        assert pytest.approx(g, abs=1e-3) == [0] * len(g)
        assert pytest.approx(g_fd, abs=1e-3) == [0] * len(g)



from __future__ import annotations

from libtbx.phil import parse

from dials.algorithms.shadowing.filter import filter_shadowed_reflections
from dials.array_family import flex
from dials.util import show_mail_handle_errors
from dials.util.command_line import Command
from dials.util.options import ArgumentParser, flatten_experiments

help_message = """

This program takes a set of experiments and predicts the reflections. The
reflections are then saved to file.

Examples::

  dials.predict models.expt

  dials.predict models.expt force_static=True

  dials.predict models.expt d_min=2.0
"""

phil_scope = parse(
    """
  output = predicted.refl
    .type = str
    .help = "The filename for the predicted reflections"

  force_static = False
    .type = bool
    .help = "For a scan varying model, force static prediction"

  ignore_shadows = True
    .type = bool
    .help = "Ignore dynamic shadowing"

  buffer_size = 0
    .type = int
    .help = "Calculate predictions within a buffer zone of n images either"
            " side of the scan"

  d_min = None
    .type = float
    .help = "Minimum d-spacing of predicted reflections"

    include scope dials.algorithms.profile_model.factory.phil_scope
""",
    process_includes=True,
)


class Script:
    """A class for running the script."""

    def __init__(self):
        """Initialise the script."""
        # The script usage
        usage = (
            "usage: dials.predict [options] [param.phil] "
            "{sequence.expt | image1.file [image2.file ...]}"
        )

        # Create the parser
        self.parser = ArgumentParser(
            usage=usage,
            phil=phil_scope,
            epilog=help_message,
            check_format=True,
            read_experiments=True,
        )

    def run(self, args=None):
        """Execute the script."""
        # Parse the command line
        params, options = self.parser.parse_args(args, show_diff_phil=True)

        # Check the number of experiments
        experiments = flatten_experiments(params.input.experiments)
        if len(experiments) == 0:
            self.parser.print_help()
            return

        predicted_all = flex.reflection_table()

        for i_expt, expt in enumerate(experiments):
            if params.buffer_size > 0:
                # Hack to make the predicter predict reflections outside of the range
                # of the scan
                scan = expt.scan
                image_range = scan.get_image_range()
                oscillation = scan.get_oscillation()
                scan.set_image_range(
                    (
                        image_range[0] - params.buffer_size,
                        image_range[1] + params.buffer_size,
                    )
                )
                scan.set_oscillation(
                    (
                        oscillation[0] - params.buffer_size * oscillation[1],
                        oscillation[1],
                    )
                )

            # Populate the reflection table with predictions
            predicted = flex.reflection_table.from_predictions(
                expt, force_static=params.force_static, dmin=params.d_min
            )
            predicted["id"] = flex.int(len(predicted), i_expt)
            predicted_all.extend(predicted)

        # if we are not ignoring shadows, look for reflections in the masked
        # region, see https://github.com/dials/dials/issues/349

        if not params.ignore_shadows:
            shadowed = filter_shadowed_reflections(
                experiments, predicted_all, experiment_goniometer=True
            )
            predicted_all = predicted_all.select(~shadowed)

        try:
            predicted_all.compute_bbox(experiments)
        except Exception:
            pass

        # Save the reflections to file
        Command.start(f"Saving {len(predicted_all)} reflections to {params.output}")
        predicted_all.as_file(params.output)
        Command.end(f"Saved {len(predicted_all)} reflections to {params.output}")


@show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

import numpy as np

from scitbx import matrix
from scitbx.array_family import flex


def plot_distance_from_ewald_sphere(experiment, reflection_table, prefix):
    """Plot distance from Ewald sphere"""

    s0 = matrix.col(experiment.beam.get_s0())
    s2 = reflection_table["s2"]
    D = flex.double(s0.length() - matrix.col(s).length() for s in s2)
    Dmean = flex.sum(D) / len(D)
    Dvar = flex.sum(flex.double([(d - Dmean) ** 2 for d in D])) / len(D)
    hist, bin_edges = np.histogram(
        D,
        bins=max(5, min(int(0.2 * len(s2)), 20)),
    )
    bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2
    plot = {
        f"{prefix}_epsilon_distribution": {
            "data": [
                (
                    {
                        "x": bin_centers.tolist(),
                        "y": hist.tolist(),
                        "type": "bar",
                    }
                )
            ],
            "layout": {
                "title": f"{prefix} epsilon distribution. <br>Mean(epsilon) = {Dmean:.2e}, Variance(epsilon) = {Dvar:.2e}",
                "xaxis": {"title": "Distance from Ewald sphere (epsilon)"},
                "yaxis": {"title": "Frequency"},
                "bargap": 0,
            },
        }
    }
    return plot


def plot_partiality(reflection_table):
    """Plot the partiality"""

    hist, bin_edges = np.histogram(
        reflection_table["partiality"],
        bins=max(5, min(int(0.2 * reflection_table.size()), 20)),
    )
    bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2
    plots_dict = {
        "partiality_distribution": {
            "data": [
                (
                    {
                        "x": bin_centers.tolist(),
                        "y": hist.tolist(),
                        "type": "bar",
                    }
                )
            ],
            "layout": {
                "title": "Partiality distribution",
                "xaxis": {"title": "Partiality"},
                "yaxis": {"title": "Frequency"},
                "bargap": 0,
            },
        }
    }
    return plots_dict



from __future__ import annotations

import logging

from libtbx.phil import parse

from dials.util import show_mail_handle_errors

logger = logging.getLogger("dials.command_line.create_profile_model")

help_message = """

This program computes the profile model from the input reflections. It then
saves a modified models.expt file with the additional profile model
information. Usually this is performed during integration; however, on some
occasions it may be desirable to compute the profile model independently.

Examples::

  dials.create_profile_model models.expt observations.refl
"""

phil_scope = parse(
    """
  subtract_background = False
    .type = bool
    .help = "Subtract background from pixel data before computing profile"
    .expert_level = 2
  output = models_with_profiles.expt
    .type = str
    .help = "The filename for the experiments"

  include scope dials.algorithms.profile_model.factory.phil_scope
  include scope dials.algorithms.spot_prediction.reflection_predictor.phil_scope
""",
    process_includes=True,
)


class Script:
    """Encapsulate the script in a class."""

    def __init__(self):
        """Initialise the script."""
        from dials.util.options import ArgumentParser

        usage = "usage: dials.create_profile_model [options] models.expt spots.refl"
        self.parser = ArgumentParser(
            usage=usage,
            epilog=help_message,
            phil=phil_scope,
            read_reflections=True,
            read_experiments=True,
            check_format=False,
        )

    def run(self, args=None):
        """Run the script."""
        from dials.algorithms.profile_model.factory import ProfileModelFactory
        from dials.array_family import flex
        from dials.util import Sorry, log
        from dials.util.command_line import Command
        from dials.util.options import reflections_and_experiments_from_files

        log.config()

        # Parse the command line
        params, options = self.parser.parse_args(args, show_diff_phil=True)
        reflections, experiments = reflections_and_experiments_from_files(
            params.input.reflections, params.input.experiments
        )
        if len(reflections) == 0 and len(experiments) == 0:
            self.parser.print_help()
            return
        if len(reflections) != 1:
            raise Sorry("exactly 1 reflection table must be specified")
        if len(experiments) == 0:
            raise Sorry("no experiments were specified")
        if ("background.mean" not in reflections[0]) and params.subtract_background:
            raise Sorry("for subtract_background need background.mean in reflections")

        reflections, _ = self.process_reference(reflections[0], params)

        # Check pixels don't belong to neighbours
        self.filter_reference_pixels(reflections, experiments)

        # Predict the reflections
        logger.info("")
        logger.info("=" * 80)
        logger.info("")
        logger.info("Predicting reflections")
        logger.info("")
        predicted = flex.reflection_table.from_predictions_multi(
            experiments,
            dmin=params.prediction.d_min,
            dmax=params.prediction.d_max,
            margin=params.prediction.margin,
            force_static=params.prediction.force_static,
            padding=params.prediction.padding,
        )

        # Match with predicted
        matched, reflections, unmatched = predicted.match_with_reference(reflections)
        assert len(matched) == len(predicted)
        assert matched.count(True) <= len(reflections)
        if matched.count(True) == 0:
            raise Sorry(
                """
        Invalid input for reference reflections.
        Zero reference spots were matched to predictions
      """
            )
        elif len(unmatched) != 0:
            logger.info("")
            logger.info("*" * 80)
            logger.info(
                "Warning: %d reference spots were not matched to predictions",
                len(unmatched),
            )
            logger.info("*" * 80)
            logger.info("")

        # Create the profile model
        experiments = ProfileModelFactory.create(params, experiments, reflections)
        for model in experiments:
            sigma_b = model.profile.sigma_b(deg=True)
            sigma_m = model.profile.sigma_m(deg=True)
            if model.profile.is_scan_varying():  # scan varying
                mean_sigma_b = sum(sigma_b) / len(sigma_b)
                mean_sigma_m = sum(sigma_m) / len(sigma_m)
                logger.info("Sigma B: %f", mean_sigma_b)
                logger.info("Sigma M: %f", mean_sigma_m)
            else:
                logger.info("Sigma B: %f", sigma_b)
                logger.info("Sigma M: %f", sigma_m)

        # Write the parameters
        Command.start(f"Writing experiments to {params.output}")
        experiments.as_file(params.output)
        Command.end(f"Wrote experiments to {params.output}")

    def process_reference(self, reference, params):
        """Load the reference spots."""
        from dials.util import Sorry

        if reference is None:
            return None, None
        assert "miller_index" in reference
        assert "id" in reference
        logger.info("Processing reference reflections")
        logger.info(" read %d strong spots", len(reference))
        mask = reference.get_flags(reference.flags.indexed)
        rubbish = reference.select(~mask)
        if mask.count(False) > 0:
            reference.del_selected(~mask)
            logger.info(" removing %d unindexed reflections", mask.count(False))
        if len(reference) == 0:
            raise Sorry(
                """
        Invalid input for reference reflections.
        Expected > %d indexed spots, got %d
      """
                % (0, len(reference))
            )
        mask = reference.get_flags(reference.flags.centroid_outlier)
        if mask.count(True) > 0:
            rubbish.extend(reference.select(mask))
            reference.del_selected(mask)
            logger.info(
                " removing %d reflections marked as centroid outliers", mask.count(True)
            )
        mask = reference["miller_index"] == (0, 0, 0)
        if mask.count(True) > 0:
            rubbish.extend(reference.select(mask))
            reference.del_selected(mask)
            logger.info(" removing %d reflections with hkl (0,0,0)", mask.count(True))
        mask = reference["id"] < 0
        if mask.count(True) > 0:
            raise Sorry(
                """
        Invalid input for reference reflections.
        %d reference spots have an invalid experiment id
      """
                % mask.count(True)
            )
        logger.info(" using %d indexed reflections", len(reference))
        logger.info(" found %d junk reflections", len(rubbish))

        if "background.mean" in reference and params.subtract_background:
            logger.info(
                " subtracting background from %d reference reflections", len(reference)
            )
            for spot in reference:
                spot["shoebox"].data -= spot["background.mean"]
        return reference, rubbish

    def filter_reference_pixels(self, reference, experiments):
        """
        Set any pixel closer to other reflections to background
        """
        modified_count = 0
        for experiment, indices in reference.iterate_experiments_and_indices(
            experiments
        ):
            subset = reference.select(indices)
            modified = subset["shoebox"].mask_neighbouring(
                subset["miller_index"],
                experiment.beam,
                experiment.detector,
                experiment.goniometer,
                experiment.scan,
                experiment.crystal,
            )
            modified_count += modified.count(True)
            reference.set_selected(indices, subset)
        logger.info(" masked neighbouring pixels in %d shoeboxes", modified_count)
        return reference


@show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

import logging

logger = logging.getLogger("dials.extensions.dispersion_spotfinder_threshold_ext")


class DispersionSpotFinderThresholdExt:
    """Extensions to do dispersion threshold."""

    name = "dispersion"

    @staticmethod
    def phil():
        from libtbx.phil import parse

        phil = parse(
            """
      gain = None
        .type = float(value_min=0.0)
        .help = "Use a flat gain map for the entire detector to act as a"
                "multiplier for the gain set by the format. Cannot be used"
                "in conjunction with lookup.gain_map parameter."

      kernel_size = 3 3
        .help = "The size of the local area around the spot in which"
                "to calculate the mean and variance. The kernel is"
                "given as a box of size (2 * nx + 1, 2 * ny + 1) centred"
                "at the pixel."
        .type = ints(size=2)
        .expert_level = 1

      sigma_background = 6
        .help = "The number of standard deviations of the index of dispersion"
                "(variance / mean) in the local area below"
                "which the pixel will be classified as background."
        .type = float
        .expert_level = 1

      sigma_strong = 3
        .help = "The number of standard deviations above the mean in the"
                "local area above which the pixel will be classified as"
                "strong."
        .type = float
        .expert_level = 1

      min_local = 2
        .help = "The minimum number of pixels under the image processing kernel"
                "that are need to do the thresholding operation. Setting the"
                "value between 2 and the total number of pixels under the"
                "kernel will force the algorithm to use that number as the"
                "minimum. If the value is less than or equal to zero, then"
                "the algorithm will use all pixels under the kernel. In"
                "effect this will add a border of pixels which are always"
                "classed as background around the edge of the image and around"
                "any masked out pixels."
        .type = int
        .expert_level = 1

      global_threshold = 0
        .type = float
        .help = "The global threshold value. Consider all pixels less than this"
                "value to be part of the background."
    """
        )
        return phil

    def __init__(self, params):
        """
        Initialise the algorithm.

        :param params: The input parameters
        """
        self.params = params

    def compute_threshold(self, image, mask, **kwargs):
        r"""
        Compute the threshold.

        :param image: The image to process
        :param mask: The pixel mask on the image
        :\*\*kwargs: Arbitrary keyword arguments
        :returns: A boolean mask showing foreground/background pixels
        """

        import libtbx

        params = self.params
        if params.spotfinder.threshold.dispersion.global_threshold is libtbx.Auto:
            params.spotfinder.threshold.dispersion.global_threshold = int(
                estimate_global_threshold(image, mask)
            )
            logger.info(
                "Setting global_threshold: %i",
                params.spotfinder.threshold.dispersion.global_threshold,
            )

        from dials.algorithms.spot_finding.threshold import DispersionThresholdStrategy

        self._algorithm = DispersionThresholdStrategy(
            kernel_size=params.spotfinder.threshold.dispersion.kernel_size,
            gain=params.spotfinder.threshold.dispersion.gain,
            mask=params.spotfinder.lookup.mask,
            n_sigma_b=params.spotfinder.threshold.dispersion.sigma_background,
            n_sigma_s=params.spotfinder.threshold.dispersion.sigma_strong,
            min_count=params.spotfinder.threshold.dispersion.min_local,
            global_threshold=params.spotfinder.threshold.dispersion.global_threshold,
        )

        return self._algorithm(image, mask)


def estimate_global_threshold(image, mask=None, plot=False):

    from scitbx import matrix
    from scitbx.array_family import flex

    n_above_threshold = flex.size_t()
    threshold = flex.double()
    for i in range(1, 20):
        g = 1.5**i
        g = int(g)
        n_above_threshold.append((image > g).count(True))
        threshold.append(g)

    # Find the elbow point of the curve, in the same manner as that used by
    # distl spotfinder for resolution method 1 (Zhang et al 2006).
    # See also dials/algorithms/spot_finding/per_image_analysis.py

    x = threshold.as_double()
    y = n_above_threshold.as_double()
    slopes = (y[-1] - y[:-1]) / (x[-1] - x[:-1])
    p_m = flex.min_index(slopes)

    x1 = matrix.col((x[p_m], y[p_m]))
    x2 = matrix.col((x[-1], y[-1]))

    gaps = flex.double()
    v = matrix.col(((x2[1] - x1[1]), -(x2[0] - x1[0]))).normalize()

    for i in range(p_m, len(x)):
        x0 = matrix.col((x[i], y[i]))
        r = x1 - x0
        g = abs(v.dot(r))
        gaps.append(g)

    p_g = flex.max_index(gaps)

    x_g_ = x[p_g + p_m]
    y_g_ = y[p_g + p_m]

    # more conservative, choose point 2 left of the elbow point
    x_g = x[p_g + p_m - 2]
    y_g = y[p_g + p_m - 2]

    if plot:
        from matplotlib import pyplot

        pyplot.figure(figsize=(16, 12))
        pyplot.scatter(threshold, n_above_threshold, marker="+")
        # for i in range(len(threshold)-1):
        #  pyplot.plot([threshold[i], threshold[-1]],
        #              [n_above_threshold[i], n_above_threshold[-1]])
        # for i in range(1, len(threshold)):
        #  pyplot.plot([threshold[0], threshold[i]],
        #              [n_above_threshold[0], n_above_threshold[i]])
        pyplot.plot([x_g, y_g], pyplot.ylim())
        pyplot.plot(
            [threshold[p_m], threshold[-1]],
            [n_above_threshold[p_m], n_above_threshold[-1]],
        )
        pyplot.plot([x_g_, threshold[-1]], [y_g_, n_above_threshold[-1]])
        pyplot.xlabel("Threshold")
        pyplot.ylabel("Number of pixels above threshold")
        pyplot.savefig("global_threshold.png")
        pyplot.clf()

    return x_g



# LIBTBX_SET_DISPATCHER_NAME dials.plot_Fo_vs_Fc
"""
Create a plot of Fo vs Fc similar to that shown by Figure 6 in
https://doi.org/10.1107/S2059798317010348

Usage: dials.plot_Fo_vs_Fc hklin=refined.mtz
"""


from __future__ import annotations

import sys
from math import sqrt

import matplotlib

from iotbx import mtz
from scitbx.array_family import flex
from scitbx.lstbx import normal_eqns, normal_eqns_solving

from dials.util import Sorry, show_mail_handle_errors
from dials.util.options import ArgumentParser

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator


class HyperbolaFit(normal_eqns.non_linear_ls, normal_eqns.non_linear_ls_mixin):
    """Fit the function y = sqrt(x^2 + a^2) by non-linear regression. There is
    just one parameter, a^2."""

    # Initial guess for the value of a^2
    a_sq0 = flex.double([1000])

    def __init__(self, x, y):
        super().__init__(n_parameters=1)
        self.x = x
        self.y = y
        self.n_data = len(self.x)
        assert len(self.y) == self.n_data
        self.restart()

    def restart(self):
        self.param = self.a_sq0.deep_copy()
        self.old_param = None

    def parameter_vector_norm(self):
        return self.param.norm()

    def build_up(self, objective_only=False):
        a_sq = self.param[0]
        model_y = flex.sqrt(flex.pow2(self.x) + a_sq)
        residuals = model_y - self.y

        self.reset()
        if objective_only:
            self.add_residuals(residuals, weights=None)
        else:
            dy_dp = 0.5 / model_y
            jacobian = flex.double(flex.grid(self.n_data, 1))
            jacobian.matrix_paste_column_in_place(dy_dp, 0)
            self.add_equations(residuals, jacobian, weights=None)

    def step_forward(self):
        self.old_param = self.param.deep_copy()
        self.param += self.step()

    def step_backward(self):
        assert self.old_param is not None
        self.param, self.old_param = self.old_param, None

    def goodness_of_fit(self):
        """Calculate various goodness of fit metrics (assumes fit has been
        performed already)"""
        a_sq = self.param[0]
        model_y = flex.sqrt(flex.pow2(self.x) + a_sq)
        resid = model_y - self.y
        resid2 = flex.pow2(resid)

        sse = flex.sum(resid2)
        sst = flex.sum(flex.pow2(model_y - flex.mean(model_y)))
        r_sq = 1 - sse / sst
        rmse = sqrt(sse / (self.n_data - 1))

        return {"SSE": sse, "R-square": r_sq, "RMSE": rmse}


class Script:
    """A class for running the script."""

    def __init__(self):
        """Initialise the script."""
        from libtbx.phil import parse

        # The phil scope
        phil_scope = parse(
            """
            hklin = None
                .type = path
                .help = "MTZ file (containing observed and calculated structure "
                        "factors)"

            Fo = F
                .type = str
                .help = "MTZ column name for Fobs"

            Fc = FC_ALL
                .type = str
                .help = "MTZ column name for Fcalc (FC_ALL from Refmac includes the "
                        "bulk solvent contribution)"

            max_Fc = 300
                .type = float
                .help = "Set plot limits to display data up to this value of Fc"

            plot_filename = Fo_vs_Fc.pdf
                .type = str
                .help = "Filename for plot"

            fit_hyperbola = True
                .type = bool
                .help = "Calculate and show the fit of a hyperbolic function given by "
                        "|Fo|^2 = |Fc|^2 + |Fe|^2, where |Fe| describes the error term "
                        "containing information about dynamic scattering and other "
                        "effects"

            show_y_eq_x = True
                .type = bool
                .help = "Plot y=x as a dashed line"
            """,
            process_includes=True,
        )

        # The script usage
        usage = "usage: dials.plot_Fo_vs_Fc hklin=refined.mtz"

        # Create the parser
        self.parser = ArgumentParser(usage=usage, phil=phil_scope, epilog=__doc__)

        self.model_fit = None

        return

    def _extract_data_from_mtz(self):
        try:
            m = mtz.object(self.params.hklin)
        except RuntimeError:
            raise Sorry(f"Could not read {self.params.hklin}")

        mad = m.as_miller_arrays_dict()
        mad = {k[-1]: v for (k, v) in mad.items()}
        fobs = mad.get(self.params.Fo)
        fc = mad.get(self.params.Fc)

        if [fobs, fc].count(None) > 0:
            raise Sorry(
                "Columns {} not found in available labels: {}".format(
                    ", ".join([self.params.Fo, self.params.Fc]),
                    ", ".join(m.column_labels()),
                )
            )

        # Find common reflections (some fobs might be missing)
        fobs, fc = fobs.common_sets(fc)

        self.fobs = fobs.data()
        self.fc = fc.amplitudes().data()

        return

    def _plot(self):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        minor_loc = MultipleLocator(10)
        ax.yaxis.set_minor_locator(minor_loc)
        ax.xaxis.set_minor_locator(minor_loc)
        ax.grid(True, which="minor")
        ax.set_axisbelow(True)
        ax.set_aspect("equal")
        ax.set_xlabel(r"$F_c$")
        ax.set_ylabel(r"$F_o$")
        ax.scatter(self.fc, self.fobs, s=1, c="indianred")

        if self.params.max_Fc:
            ax.set_xlim((0, self.params.max_Fc))
            ax.set_ylim((0, self.params.max_Fc))

        if self.params.show_y_eq_x:
            ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="0.0", linewidth=0.8)

        if self.model_fit:
            x = flex.double_range(0, int(ax.get_xlim()[1]))
            y = self.model_fit(x)
            ax.plot(x, y, c="0.0", linewidth=0.8)

        print(f"Saving plot to {self.params.plot_filename}")
        plt.savefig(self.params.plot_filename)

    def run(self, args=None):
        """Execute the script."""

        # Parse the command line
        self.params, _ = self.parser.parse_args(args, show_diff_phil=True)

        if self.params.hklin is None:
            self.parser.print_help()
            sys.exit()

        self._extract_data_from_mtz()

        if self.params.fit_hyperbola:
            # fit by NLLS Levenberg Marquardt algorithm
            hyperbola_fit = HyperbolaFit(self.fc, self.fobs)
            hyperbola_fit.restart()
            normal_eqns_solving.levenberg_marquardt_iterations(
                hyperbola_fit,
                track_all=True,
                gradient_threshold=1e-8,
                step_threshold=1e-8,
                tau=1e-4,
                n_max_iterations=200,
            )
            intercept = hyperbola_fit.param[0]

            print("Model fit described by the formula: |Fo|^2 = sqrt(|Fc|^2 + |Fe|^2)")
            print(f"where |Fe| = {sqrt(intercept):.5f}\n")

            print("Goodness of fit:")
            gof = hyperbola_fit.goodness_of_fit()
            print(f"SSE: {gof['SSE']:.5g}")
            print(f"R-square: {gof['R-square']:.5f}")
            print(f"RMSE: {gof['RMSE']:.2f}")
            print()

            # Set the model_fit function using the determined intercept
            def hyperbola(x, c):
                return flex.sqrt(flex.pow2(x) + c)

            from functools import partial

            self.model_fit = partial(hyperbola, c=intercept)

        if self.params.plot_filename:
            self._plot()

        return


@show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

from cctbx import uctbx
from scitbx.array_family import flex


def quasi_normalisation(intensities):
    """Quasi-normalisation of the input intensities.

    Args:
        intensities (cctbx.miller.array): The intensities to be normalised.

    Returns:
        cctbx.miller.array: The normalised intensities.
    """
    # handle negative reflections to minimise effect on mean I values.
    work = intensities.deep_copy()
    work.data().set_selected(work.data() < 0.0, 0.0)

    # set up binning objects
    if work.size() > 20000:
        n_refl_shells = 20
    elif work.size() > 15000:
        n_refl_shells = 15
    else:
        n_refl_shells = 10
    d_star_sq = work.d_star_sq().data()
    d_star_sq_max = flex.max(d_star_sq)
    d_star_sq_min = flex.min(d_star_sq)
    span = d_star_sq_max - d_star_sq_min
    d_star_sq_max += span * 1e-6
    d_star_sq_min -= span * 1e-6
    d_star_sq_step = (d_star_sq_max - d_star_sq_min) / n_refl_shells
    work.setup_binner_d_star_sq_step(
        d_min=uctbx.d_star_sq_as_d(d_star_sq_min),  # cctbx/cctbx_project#588
        d_max=uctbx.d_star_sq_as_d(d_star_sq_max),  # cctbx/cctbx_project#588
        d_star_sq_step=d_star_sq_step,
        auto_binning=False,
    )
    normalisations = work.intensity_quasi_normalisations()
    return intensities.customized_copy(
        data=(intensities.data() / normalisations.data()),
        sigmas=(intensities.sigmas() / normalisations.data()),
    )



from __future__ import annotations

from scitbx.array_family import flex


class Parameter:
    """A class to help formalise what a parameter is.

    A Parameter must have a numerical value (either a length or an angle). It may
    also have a vector axis which provides context for what that number means.

    Together the values and axes of a set of parameters can be used to compose
    the state of a model. For example, the value might be a rotation angle, with
    the axis of rotation providing the context.

    A slot is also provided for storage of the estimated standard deviation of
    the value, which may be of use in future. Whenever the parameter
    value is set, the esd is reset to None. So this must be set separately, and
    after the parameter value if it is required.
    """

    def __init__(self, value, axis=None, ptype=None, name="Parameter"):
        """Initialisation for the Parameter class.

        Args:
            value (float): The initial parameter value.
            axis: A 3-element vector giving the axis of action of the parameter.
                Defaults to None.
            ptype (str): A string defining the parameter type.
            name (str): A string defining the parameter name.
        """
        self._value = value
        self._esd = None
        self._axis = axis
        self._ptype = ptype
        self._name = name
        self._fixed = False

        return

    @property
    def value(self):
        """The floating point parameter value."""
        return self._value

    @value.setter
    def value(self, val):
        self._value = val
        self._esd = None

    @property
    def name(self):
        """A string identifying the parameter name."""
        return self._name

    @property
    def esd(self):
        """The floating point parameter estimated standard deviation value."""
        return self._esd

    @esd.setter
    def esd(self, esd):
        self._esd = esd

    @property
    def param_type(self):
        """A string identifying the parameter type."""
        return self._ptype

    @property
    def axis(self):
        """A 3D column vector giving the direction of action of the parameter."""
        return self._axis

    def get_fixed(self):
        """Return a bool describing whether the parameter is fixed or not."""
        return self._fixed

    def fix(self):
        """Fix the parameter"""
        self._fixed = True

    def unfix(self):
        """Unfix (free) the parameter"""
        self._fixed = False

    def __str__(self):
        """Provide a descriptive string representation of the Parameter class"""

        msg = "Parameter " + self.name + ":\n"
        try:
            msg += "    Type: " + self.param_type + "\n"
        except TypeError:
            msg += "    Type: " + str(self.param_type) + "\n"
        try:
            msg += "    Axis: (%5.3f, %5.3f, %5.3f)" % tuple(self.axis) + "\n"
        except TypeError:
            msg += "    Axis: " + str(self.axis) + "\n"
        msg += f"    Value: {self.value:5.3f}" + "\n"
        try:
            msg += f"    Sigma: {self.esd:5.3f}" + "\n"
        except TypeError:
            msg += "    Sigma: " + str(self.esd) + "\n"

        return msg


class ModelParameterisation:
    """An abstract interface for the parameterisation of a model.

    Parameterisation of experimental objects, such as the detector, the beam,
    etc. should adhere to this interface in order to compose their state from
    their parameters, access their parameters, and derivatives of their state wrt
    their parameters, taking into account whether particular parameters are fixed
    or free.

    It is possible to parameterise a model with multiple states. The first such
    example is a detector with multiple panels. Each panel has its own matrix
    describing its geometrical 'state'. One set of parameters is used to compose
    all states and calculate all derivatives of these states.
    """

    def __init__(
        self, model, initial_state, param_list, experiment_ids, is_multi_state=False
    ):
        """Initialisation for the ModelParameterisation abstract base class.

        Args:
            model: The experimental model to be parameterised.
            initial_state: The initial value of the state of interest of the model.
            param_list (list): A list of Parameter objects.
            experiment_ids (list): A list of integer experiment IDs that this
                parameterisation affects.
            is_multi_state (bool): A flag to indicate whether the parameterisation
                is for a multi-state model.
        """
        assert isinstance(param_list, list)
        self._initial_state = initial_state
        self._model = model
        self._param = list(param_list)
        self._total_len = len(self._param)
        self._num_free = None
        self._dstate_dp = [None] * len(param_list)
        self._is_multi_state = is_multi_state
        if is_multi_state:
            self._multi_state_derivatives = [None] * len(model)
        self._exp_ids = experiment_ids
        self._null_state = self.get_state() * 0.0

        # An attribute that can be set to name the model being parameterised
        self.model_identifier = ""
        return

    def is_multi_state(self):
        """Query whether this is a multi-state parameterisation or not (e.g. True
        for a multi-panel detector parameterisation)"""

        return self._is_multi_state

    def num_free(self):
        """Get the number of free parameters"""

        if self._num_free is None:
            self._num_free = sum(not x.get_fixed() for x in self._param)
        return self._num_free

    def num_total(self):
        """Get the total number of parameters, both fixed and free"""
        return self._total_len

    def get_experiment_ids(self):
        """Get the list of experiment IDs of parameterised experiments."""
        return self._exp_ids

    def get_model(self):
        """Get the model that is parameterised"""
        return self._model

    def compose(self):
        """Compose the current model state.

        Using the initial state and the current parameter list, compose the current
        state of the model. Also calculate the derivatives of the state wrt each
        parameter in the list. Intended to be called automatically once parameters
        are updated, e.g. at the end of each refinement cycle.
        """

        # Specific methods for each model are defined by derived classes.
        raise NotImplementedError()

    def get_params(self, only_free=True):
        """Get the internal list of parameters. It is intended that this function
        be used for reporting parameter attributes, not for modifying them.

        Args:
            only_free (bool): Whether fixed parameters should be filtered from the
                returned list, or all parameters returned.
        """

        if only_free:

            return [x for x in self._param if not x.get_fixed()]

        else:
            return list(self._param)

    def get_param_vals(self, only_free=True):
        """Get the values of the internal list of parameters as a sequence of
        floats.

        Args:
            only_free (bool): Whether fixed parameters should have their values
            filtered from the returned list, or all parameter values returned.
        """

        if only_free:

            return [x.value for x in self._param if not x.get_fixed()]

        else:
            return [x.value for x in self._param]

    def get_param_names(self, only_free=True):
        """Get the list of names of the parameters.

        Args:
            only_free (bool): Whether fixed parameters should have their names
            filtered from the returned list, or all parameter names returned.
        """

        if only_free:
            return [x.name for x in self._param if not x.get_fixed()]

        else:
            return [x.name for x in self._param]

    def set_param_vals(self, vals):
        """Set the values of the internal list of parameters from a sequence of
        floats.

        Args:
            vals (list): A list of floating point parameter values, equal in length
                to the number of free parameters.
        """

        assert len(vals) == self.num_free()

        v = iter(vals)
        for p in self._param:
            if not p.get_fixed():  # only set the free parameters
                p.value = next(v)
                p.esd = None

        # compose with the new parameter values
        self.compose()

        return

    def set_param_esds(self, esds):
        """Set the estimated standard deviations of the internal list of parameters
        from a sequence of floats.

        Args:
            esds (list): A list of floating point parameter esd values, equal in
                length to the number of free parameters.
        """

        assert len(esds) == self.num_free()

        v = iter(esds)
        for p in self._param:
            if not p.get_fixed():  # only set the free parameters
                p.esd = next(v)

        return

    def get_fixed(self):
        """Get the list determining whether each parameter is fixed or not"""

        return [p.get_fixed() for p in self._param]

    def set_fixed(self, fix):
        """Set parameters to be fixed or free.

        Args:
            fix (list): A list of bools, determining whether each parameter is
                fixed or free.
        """

        assert len(fix) == len(self._param)

        for f, p in zip(fix, self._param):
            if f:
                p.fix()
            else:
                p.unfix()

        # reset the cached number of free parameters
        self._num_free = None

        return

    def get_state(self, multi_state_elt=None):
        """Get the current state of the model under parameterisation.

        This is required, for example, by the calculation of finite
        difference gradients.

        Args:
            multi_state_elt (bool): For a multi-state parameterisation, the
                requested state is selected by an integer array index. Defaults to
                None.
        """

        # To be implemented by the derived class, where it is clear what aspect
        # of the model under parameterisation is considered its state. The
        # type of this result should match the type of one element of the return
        # value of get_ds_dp.
        raise NotImplementedError()

    def get_ds_dp(self, only_free=True, multi_state_elt=None, use_none_as_null=False):
        """Get a list of derivatives of the state wrt each parameter.

        This is returned as a list in the same order as the internal list of
        parameters.

        The internal list of derivatives may use None for null elements. By default
        these are converted to the null state, but optionally these may remain None
        to detect them easier and avoid doing calculations on null elements.

        Args:
            only_free (bool): whether the derivatives with respect to fixed
                parameters are omitted from the returned list. Otherwise a list for
                all parameters is returned, with null values for the fixed
                parameters.
            multi_state_elt (bool): For a multi-state parameterisation, the
                requested derivative of the state is selected by an integer array
                index. Defaults to None.
            use_none_as_null (bool): See method description. Defaults to False.
        """

        if use_none_as_null:
            null = None
        else:
            null = self._null_state

        if self._is_multi_state:
            if multi_state_elt is None:
                raise ValueError(
                    "Must provide multi_state_elt for a multi-state parameterisation"
                )
            else:
                # copy results for the state of interest
                self._dstate_dp = self._multi_state_derivatives[multi_state_elt]

        if only_free:
            grads = [
                ds_dp
                for ds_dp, p in zip(self._dstate_dp, self._param)
                if not p.get_fixed()
            ]
        else:
            grads = self._dstate_dp

        return [null if e is None else e for e in grads]

    def calculate_state_uncertainties(self, var_cov):
        """Given a variance-covariance array for the parameters of this model,
        propagate those estimated errors into the uncertainties of the model state

        Args:
            var_cov: A variance-covariance matrix for the parameters of this model.
        """

        grads = []
        if self._is_multi_state:
            i = 0
            while True:
                try:
                    grads.append(self.get_ds_dp(multi_state_elt=i))
                    i += 1
                except IndexError:
                    break
        else:
            grads.append(self.get_ds_dp())

        if len(grads[0]) == 0:
            return

        # the jacobian is the m*n matrix of partial derivatives of the m state
        # elements wrt the n parameters
        state_covs = []
        for grads_one_state in grads:
            jacobian_t = flex.double(e for g in grads_one_state for e in g)
            jacobian_t.reshape(
                flex.grid(len(grads_one_state), len(grads_one_state[0].elems))
            )

            # propagation of errors takes the variance-covariance matrix of parameters,
            # along with the jacobian mapping changes in parameter values to changes
            # in the model state elements, to calculate an approximate variance-
            # covariance matrix of the state elements. That is, state_cov is the
            # matrix product: jacobian * var_cov * jacobian_t
            tmp = var_cov.matrix_multiply(jacobian_t)
            state_cov = jacobian_t.matrix_transpose_multiply(tmp).as_scitbx_matrix()
            state_covs.append(state_cov)

        return state_covs

    def set_state_uncertainties(self, var_cov, multi_state_elt=None):
        """Send the calculated variance-covariance matrix for model state elements
        back to the model for storage alongside the model state, and potentially
        use in further propagation of error calculations.

        Args:
            var_cov: A variance-covariance matrix for the parameters of this model.
            multi_state_elt (bool): If the parameterisation is multi-state, a
                integer index is required to select which state this variance-
                covariance matrix refers to.
        """

        # To be implemented by the derived class, where it is clear what aspect
        # of the model under parameterisation is considered its state.
        pass



"""
Identify connected regions of missing reflections in the asymmetric unit.

This is achieved by first generating the complete set of possible miller indices,
then performing connected components analysis on a graph of nearest neighbours in
the list of missing reflections.

Examples::

  dials.missing_reflections integrated.expt integrated.refl

  dials.missing_reflections scaled.expt scaled.refl min_component_size=10
"""

from __future__ import annotations

import io
import logging
import sys
from typing import List

import libtbx.phil
from cctbx import uctbx

import dials.util.log
from dials.report.analysis import scaled_data_as_miller_array
from dials.util import missing_reflections, tabulate
from dials.util.filter_reflections import filtered_arrays_from_experiments_reflections
from dials.util.options import ArgumentParser, flatten_experiments, flatten_reflections
from dials.util.version import dials_version

logger = logging.getLogger("dials.missing_reflections")

phil_scope = libtbx.phil.parse(
    """
    min_component_size = 0
      .type = int(value_min=0)
      .help = "Only show connected regions larger than or equal to this."
    d_min = None
      .type = float(value_min=0)
    d_max = None
      .type = float(value_min=0)
    """
)


@dials.util.show_mail_handle_errors()
def run(args: List[str] = None, phil: libtbx.phil.scope = phil_scope) -> None:
    usage = "dials.missing_reflections [options] scaled.expt scaled.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=__doc__,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=False)

    # Configure the logging.
    dials.util.log.config(options.verbose)
    logger.info(dials_version())

    # Log the difference between the PHIL scope definition and the active PHIL scope,
    # which will include the parsed user inputs.
    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)

    experiments = flatten_experiments(params.input.experiments)
    reflections = flatten_reflections(params.input.reflections)

    if not experiments or not reflections:
        parser.print_help()
        return
    if len(reflections) != 1 and len(experiments) != len(reflections):
        sys.exit("Number of experiments must equal the number of reflection tables")

    from dials.util.multi_dataset_handling import (
        assign_unique_identifiers,
        parse_multiple_datasets,
    )

    reflections = parse_multiple_datasets(reflections)
    experiments, reflections = assign_unique_identifiers(experiments, reflections)

    if all("inverse_scale_factor" in refl for refl in reflections):
        # Assume all arrays have been scaled
        miller_array = scaled_data_as_miller_array(
            reflections, experiments, anomalous_flag=False
        )
    else:
        # Else get the integrated intensities
        miller_arrays = filtered_arrays_from_experiments_reflections(
            experiments,
            reflections,
        )
        miller_array = miller_arrays[0]
        for ma in miller_arrays[1:]:
            miller_array = miller_array.concatenate(ma)

    if params.d_min or params.d_max:
        miller_array = miller_array.resolution_filter(
            d_min=params.d_min, d_max=params.d_max
        )

    # Print overall summary of input miller array
    s = io.StringIO()
    ma_unique = miller_array.unique_under_symmetry()
    ma_unique.show_comprehensive_summary(f=s)
    logger.info(f"\n{s.getvalue()}")

    # Get the regions of missing reflections
    complete_set, unique_ms = missing_reflections.connected_components(miller_array)
    unique_ms = [ms for ms in unique_ms if ms.size() >= params.min_component_size]

    # Print some output for user
    if len(unique_ms):
        logger.info(
            "The following connected regions of missing reflections have been identified:"
        )
        n_expected = complete_set.size()
        rows = []
        for ms in unique_ms:
            d_max, d_min = (uctbx.d_star_sq_as_d(ds2) for ds2 in ms.min_max_d_star_sq())
            rows.append(
                [
                    ms.size(),
                    f"{100 * ms.size() / n_expected:.1f}",
                    f"{d_max:.2f}-{d_min:.2f}",
                ]
            )
        logger.info(
            tabulate(
                rows, headers=["# reflections", "% missing", "Resolution range (Å)"]
            )
        )
    else:
        logger.info("No connected regions of missing reflections identified")


if __name__ == "__main__":
    run()



from __future__ import annotations

import math

import pytest

from . import geometry_phil


def test():
    from cctbx.sgtbx import space_group, space_group_symbols
    from dxtbx.model.experiment_list import Experiment, ExperimentList
    from libtbx.phil import parse
    from scitbx.array_family import flex

    from dials.algorithms.refinement.parameterisation.beam_parameters import (
        BeamParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalOrientationParameterisation,
        CrystalUnitCellParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.detector_parameters import (
        DetectorParameterisationSinglePanel,
    )
    from dials.algorithms.refinement.parameterisation.goniometer_parameters import (
        GoniometerParameterisation,
    )

    #### Import model parameterisations
    from dials.algorithms.refinement.parameterisation.prediction_parameters import (
        XYPhiPredictionParameterisation,
    )
    from dials.algorithms.refinement.prediction.managed_predictors import (
        ScansExperimentsPredictor,
        ScansRayPredictor,
    )

    ##### Imports for reflection prediction
    from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

    ##### Import model builder
    from .setup_geometry import Extract

    #### Create models

    overrides = """geometry.parameters.crystal.a.length.range = 10 50
  geometry.parameters.crystal.b.length.range = 10 50
  geometry.parameters.crystal.c.length.range = 10 50"""

    master_phil = parse(geometry_phil)

    models = Extract(master_phil, overrides)

    mydetector = models.detector
    mygonio = models.goniometer
    mycrystal = models.crystal
    mybeam = models.beam

    # Build a mock scan for a 72 degree sequence
    sequence_range = (0.0, math.pi / 5.0)
    from dxtbx.model import ScanFactory

    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 720),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(720)),
        deg=True,
    )

    #### Create parameterisations of these models
    det_param = DetectorParameterisationSinglePanel(mydetector)
    s0_param = BeamParameterisation(mybeam, mygonio)
    xlo_param = CrystalOrientationParameterisation(mycrystal)
    xluc_param = CrystalUnitCellParameterisation(mycrystal)
    gon_param = GoniometerParameterisation(mygonio, mybeam)

    # Create an ExperimentList
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal,
            imageset=None,
        )
    )

    #### Unit tests

    # Build a prediction parameterisation
    pred_param = XYPhiPredictionParameterisation(
        experiments,
        detector_parameterisations=[det_param],
        beam_parameterisations=[s0_param],
        xl_orientation_parameterisations=[xlo_param],
        xl_unit_cell_parameterisations=[xluc_param],
        goniometer_parameterisations=[gon_param],
    )

    # Generate reflections
    resolution = 2.0
    index_generator = IndexGenerator(
        mycrystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()

    # Predict rays within the sequence range
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    obs_refs = ray_predictor(indices)

    # Take only those rays that intersect the detector
    intersects = ray_intersection(mydetector, obs_refs)
    obs_refs = obs_refs.select(intersects)

    # Make a reflection predictor and re-predict for all these reflections. The
    # result is the same, but we gain also the flags and xyzcal.px columns
    ref_predictor = ScansExperimentsPredictor(experiments)
    obs_refs["id"] = flex.int(len(obs_refs), 0)
    obs_refs = ref_predictor(obs_refs)

    # Set 'observed' centroids from the predicted ones
    obs_refs["xyzobs.mm.value"] = obs_refs["xyzcal.mm"]

    # Invent some variances for the centroid positions of the simulated data
    im_width = 0.1 * math.pi / 180.0
    px_size = mydetector[0].get_pixel_size()
    var_x = flex.double(len(obs_refs), (px_size[0] / 2.0) ** 2)
    var_y = flex.double(len(obs_refs), (px_size[1] / 2.0) ** 2)
    var_phi = flex.double(len(obs_refs), (im_width / 2.0) ** 2)
    obs_refs["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)

    # use a ReflectionManager to exclude reflections too close to the spindle
    from dials.algorithms.refinement.reflection_manager import ReflectionManager

    refman = ReflectionManager(obs_refs, experiments, outlier_detector=None)
    refman.finalise()

    # Redefine the reflection predictor to use the type expected by the Target class
    ref_predictor = ScansExperimentsPredictor(experiments)

    # keep only those reflections that pass inclusion criteria and have predictions
    reflections = refman.get_matches()

    # get analytical gradients
    an_grads = pred_param.get_gradients(reflections)

    # get finite difference gradients
    p_vals = pred_param.get_param_vals()
    deltas = [1.0e-7] * len(p_vals)

    for i, delta in enumerate(deltas):
        val = p_vals[i]

        p_vals[i] -= delta / 2.0
        pred_param.set_param_vals(p_vals)

        ref_predictor(reflections)

        rev_state = reflections["xyzcal.mm"].deep_copy()

        p_vals[i] += delta
        pred_param.set_param_vals(p_vals)

        ref_predictor(reflections)

        fwd_state = reflections["xyzcal.mm"].deep_copy()
        p_vals[i] = val

        fd = fwd_state - rev_state
        x_grads, y_grads, phi_grads = fd.parts()
        x_grads /= delta
        y_grads /= delta
        phi_grads /= delta

        # compare with analytical calculation
        assert x_grads == pytest.approx(an_grads[i]["dX_dp"], abs=5.0e-6)
        assert y_grads == pytest.approx(an_grads[i]["dY_dp"], abs=5.5e-6)
        assert phi_grads == pytest.approx(an_grads[i]["dphi_dp"], abs=5.0e-6)

    # return to the initial state
    pred_param.set_param_vals(p_vals)



from __future__ import annotations

from dxtbx.model.experiment_list import Experiment, ExperimentList

from dials.algorithms.indexing.indexer import Indexer


class IndexerKnownOrientation(Indexer):
    def __init__(self, reflections, experiments, params, known_orientations):
        self.known_orientations = known_orientations
        super().__init__(reflections, experiments, params)

    def find_lattices(self):
        experiments = ExperimentList()
        for cm, expt in zip(self.known_orientations, self.experiments):
            # indexer expects crystals to be in primitive setting
            space_group = cm.get_space_group()
            cb_op_to_primitive = (
                space_group.info().change_of_basis_op_to_primitive_setting()
            )
            cm = cm.change_basis(cb_op_to_primitive)
            experiments.append(
                Experiment(
                    imageset=expt.imageset,
                    beam=expt.beam,
                    detector=expt.detector,
                    goniometer=expt.goniometer,
                    scan=expt.scan,
                    crystal=cm,
                    identifier=expt.identifier,
                )
            )
        return experiments



from __future__ import annotations

import json
import math
import os
import pathlib
import shutil

import procrunner
import pytest

from dxtbx.serialize import load

from dials.algorithms.integration.processor import _average_bbox_size
from dials.array_family import flex


def test_basic_integrate(dials_data, tmpdir):
    # Call dials.integrate

    exp = load.experiment_list(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    )
    exp[0].identifier = "foo"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))

    exp.as_json(tmpdir.join("modified_input.json"))

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "modified_input.json",
            "profile.fitting=False",
            "integration.integrator=3d",
            "prediction.padding=0",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr

    experiments = load.experiment_list(tmpdir / "integrated.expt")
    assert experiments[0].identifier == "foo"

    table = flex.reflection_table.from_file(tmpdir / "integrated.refl")
    mask = table.get_flags(table.flags.integrated, all=False)
    assert len(table) == 1666
    assert mask.count(True) == 1666
    assert "id" in table
    for row in table.rows():
        assert row["id"] == 0

    assert dict(table.experiment_identifiers()) == {0: "foo"}

    originaltable = table

    tmpdir.join("integrated.refl").remove()

    tmp_path = pathlib.Path(tmpdir)
    for i in range(1, 10):
        source = dials_data("centroid_test_data", pathlib=True) / f"centroid_000{i}.cbf"
        destination = tmp_path / f"centroid_001{i}.cbf"
        try:
            destination.symlink_to(source)
        except OSError:
            shutil.copyfile(source, destination)

    with dials_data("centroid_test_data", pathlib=True).joinpath(
        "experiments.json"
    ).open("r") as fh:
        j = json.load(fh)
    assert j["scan"][0]["image_range"] == [1, 9]
    j["scan"][0]["image_range"] = [11, 19]
    assert j["scan"][0]["oscillation"] == [0.0, 0.2]
    j["scan"][0]["oscillation"] = [360.0, 0.2]
    j["experiment"][0]["identifier"] = "bar"
    with tmpdir.join("models.expt").open("w") as fh:
        json.dump(j, fh)

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    exp = load.experiment_list(tmpdir.join("models.expt"))
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))

    exp.as_json(tmpdir.join("models.expt"))

    # Call dials.integrate
    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "models.expt",
            "profile.fitting=False",
            "integration.integrator=3d",
            "prediction.padding=0",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr

    experiments = load.experiment_list(tmpdir / "integrated.expt")
    assert experiments[0].identifier == "bar"

    table = flex.reflection_table.from_file(tmpdir / "integrated.refl")
    assert dict(table.experiment_identifiers()) == {0: "bar"}
    mask1 = table.get_flags(table.flags.integrated, all=False)
    assert len(table) == 1666
    assert mask1.count(True) == 1666
    mask2 = originaltable.get_flags(table.flags.integrated, all=False)
    assert mask1.all_eq(mask2)
    t1 = table.select(mask1)
    t2 = originaltable.select(mask1)
    Cal_P1 = t1["xyzcal.mm"].parts()[2]
    Cal_Z1 = t1["xyzcal.px"].parts()[2]
    Obs_Z1 = t1["xyzobs.px.value"].parts()[2]
    # Obs_P1 = t1['xyzobs.mm.value'].parts()[2]
    Cal_Z2 = t2["xyzcal.px"].parts()[2]
    Cal_P2 = t2["xyzcal.mm"].parts()[2]
    Obs_Z2 = t2["xyzobs.px.value"].parts()[2]
    # Obs_P2 = t2['xyzobs.mm.value'].parts()[2]
    diff_I = t1["intensity.sum.value"] - t2["intensity.sum.value"]
    diff_Cal_Z = Cal_Z1 - (Cal_Z2 + 10)
    diff_Obs_Z = Obs_Z1 - (Obs_Z2 + 10)
    diff_Cal_P = Cal_P1 - (Cal_P2 + 2 * math.pi)
    # diff_Obs_P = Obs_P1 - (Obs_P2 + 2*math.pi)
    assert flex.abs(diff_I).all_lt(1e-7)
    assert flex.abs(diff_Cal_Z).all_lt(1e-7)
    assert flex.abs(diff_Cal_P).all_lt(1e-7)
    assert flex.abs(diff_Obs_Z).all_lt(1e-7)
    # assert(flex.abs(diff_Obs_P).all_lt(1e-7))


@pytest.mark.parametrize(
    ("block_size", "block_units"),
    [(None, None), (1, "degrees"), (2, "frames"), (1, "frames")],
)
def test_basic_blocking_options(dials_data, tmp_path, block_size, block_units):
    exp = load.experiment_list(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    )
    exp[0].identifier = "foo"
    exp.as_json(tmp_path / "modified_input.json")

    args = ["dials.integrate", "modified_input.json", "nproc=2"]
    if block_size:
        args.append(f"block.size={block_size}")
    if block_units:
        args.append(f"block.units={block_units}")

    result = procrunner.run(args, working_directory=tmp_path)
    assert not result.returncode and not result.stderr


def test_basic_threaded_integrate(dials_data, tmp_path):
    """Test the threaded integrator on single imageset data."""

    expts = dials_data("centroid_test_data", pathlib=True) / "indexed.expt"
    refls = dials_data("centroid_test_data", pathlib=True) / "indexed.refl"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    exp = load.experiment_list(expts)
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    exp.as_json(tmp_path / "trusted_range_patch.expt")

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "integration.integrator=3d_threaded",
            "background.algorithm=glm",
            "njobs=2",
            "nproc=2",
            refls,
            "trusted_range_patch.expt",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("integrated.refl").is_file()
    assert tmp_path.joinpath("integrated.expt").is_file()

    table = flex.reflection_table.from_file(tmp_path / "integrated.refl")
    assert table.size() == 3526
    assert set(table["id"]) == {0}
    assert table.select(table["id"] == 0).size() == 3526


def test_basic_integrate_output_integrated_only(dials_data, tmpdir):

    exp = load.experiment_list(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    )
    exp[0].identifier = "bar"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))

    exp.as_json(tmpdir.join("modified_input.json"))

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "modified_input.json",
            "profile.fitting=False",
            "integration.integrator=3d",
            "output_unintegrated_reflections=False",
            "prediction.padding=0",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr

    experiments = load.experiment_list(tmpdir / "integrated.expt")
    assert experiments[0].identifier == "bar"

    table = flex.reflection_table.from_file(tmpdir / "integrated.refl")
    mask = table.get_flags(table.flags.integrated, all=False)
    assert len(table) == 1666
    assert mask.count(False) == 0
    assert "id" in table
    for row in table.rows():
        assert row["id"] == 0

    assert dict(table.experiment_identifiers()) == {0: "bar"}


def test_integration_with_sampling(dials_data, tmpdir):

    exp = load.experiment_list(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    )
    exp[0].identifier = "foo"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    # exp.as_json(tmp_path / "trusted_range_patch.expt")

    exp.as_json(tmpdir / "modified_input.json")

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "modified_input.json",
            "profile.fitting=False",
            "sampling.integrate_all_reflections=False",
            "sampling.random_seed=42",
            "prediction.padding=0",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    experiments = load.experiment_list(tmpdir / "integrated.expt")
    assert experiments[0].identifier == "foo"

    table = flex.reflection_table.from_file(tmpdir / "integrated.refl")

    # account for random number generator in sampling
    assert len(table) == 839

    assert dict(table.experiment_identifiers()) == {0: "foo"}


def test_integration_with_sample_size(dials_data, tmpdir):
    exp = load.experiment_list(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    )
    exp[0].identifier = "foo"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    # exp.as_json(tmp_path / "trusted_range_patch.expt")

    exp.as_json(tmpdir / "modified_input.json")

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "modified_input.json",
            "profile.fitting=False",
            "sampling.integrate_all_reflections=False",
            "sampling.random_seed=42",
            "sampling.minimum_sample_size=500",
            "prediction.padding=0",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    experiments = load.experiment_list(tmpdir / "integrated.expt")
    assert experiments[0].identifier == "foo"
    table = flex.reflection_table.from_file(tmpdir / "integrated.refl")
    assert len(table) == 415
    assert dict(table.experiment_identifiers()) == {0: "foo"}


def test_imageset_id_output_with_multi_sweep(dials_data, tmp_path):
    """Test that imageset ids are correctly output for multi-sweep integration."""
    # Just integrate 15 images for each sweep

    images1 = dials_data("l_cysteine_dials_output", pathlib=True) / "l-cyst_01_000*.cbf"
    images2 = dials_data("l_cysteine_dials_output", pathlib=True) / "l-cyst_02_000*.cbf"

    result = procrunner.run(
        ["dials.import", images1, images2], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        ["dials.find_spots", tmp_path / "imported.expt", "nproc=1"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        ["dials.index", tmp_path / "imported.expt", tmp_path / "strong.refl"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            tmp_path / "indexed.expt",
            tmp_path / "indexed.refl",
            "profile.fitting=False",
            "gaussian_rs.min_spots.overall=0",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    table = flex.reflection_table.from_file(tmp_path / "integrated.refl")
    assert set(table["imageset_id"]) == {0, 1}
    # check that we have approx 50% in each
    n0 = (table["imageset_id"] == 0).count(True)
    n = table.size()
    n1 = (table["imageset_id"] == 1).count(True)
    assert (n0 / n > 0.4) and (n0 / n < 0.6)
    assert (n1 / n > 0.4) and (n1 / n < 0.6)

    # now try again with adding unintegrated reflections
    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            tmp_path / "indexed.expt",
            tmp_path / "indexed.refl",
            "profile.fitting=False",
            "gaussian_rs.min_spots.overall=0",
            "output_unintegrated_reflections=False",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    table = flex.reflection_table.from_file(tmp_path / "integrated.refl")
    assert set(table["imageset_id"]) == {0, 1}
    # check that we have approx 50% in each
    n0 = (table["imageset_id"] == 0).count(True)
    n = table.size()
    n1 = (table["imageset_id"] == 1).count(True)
    assert (n0 / n > 0.4) and (n0 / n < 0.6)
    assert (n1 / n > 0.4) and (n1 / n < 0.6)


def test_basic_integration_with_profile_fitting(dials_data, tmpdir):
    expts = dials_data("centroid_test_data", pathlib=True) / "indexed.expt"
    refls = dials_data("centroid_test_data", pathlib=True) / "indexed.refl"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    exp = load.experiment_list(expts)
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    exp.as_json(tmpdir.join("trusted_range_patch.expt"))

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "trusted_range_patch.expt",
            refls,
            "profile.fitting=True",
            "sampling.integrate_all_reflections=False",
            "sampling.random_seed=42",
            "sampling.minimum_sample_size=500",
            "prediction.padding=0",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    table = flex.reflection_table.from_file(tmpdir / "integrated.refl")
    assert len(table) == 436

    prf = table.get_flags(table.flags.integrated_prf)
    zero = table["intensity.prf.value"] == 0.0
    prf_and_zero = prf & zero
    assert prf_and_zero.count(True) == 0


def test_multi_sweep(dials_regression, tmpdir):
    expts = os.path.join(
        dials_regression, "integration_test_data", "multi_sweep", "experiments.json"
    )

    experiments = load.experiment_list(expts)
    for i, expt in enumerate(experiments):
        expt.identifier = str(100 + i)

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    panel = experiments[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))

    experiments.as_json(tmpdir / "modified_input.json")

    refls = os.path.join(
        dials_regression, "integration_test_data", "multi_sweep", "indexed.pickle"
    )

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "modified_input.json",
            refls,
            "prediction.padding=0",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert (tmpdir / "integrated.refl").check()
    assert (tmpdir / "integrated.expt").check()

    experiments = load.experiment_list(tmpdir / "integrated.expt")
    for i, expt in enumerate(experiments):
        assert expt.identifier == str(100 + i)

    table = flex.reflection_table.from_file(tmpdir / "integrated.refl")
    assert len(table) == 3530
    assert dict(table.experiment_identifiers()) == {0: "100", 1: "101"}

    # Check the results
    T1 = table[:1765]
    T2 = table[1765:]
    ID1 = list(set(T1["id"]))
    ID2 = list(set(T2["id"]))
    assert len(ID1) == 1
    assert len(ID2) == 1
    assert ID1[0] == 0
    assert ID2[0] == 1
    I1 = T1["intensity.prf.value"]
    I2 = T2["intensity.prf.value"]
    F1 = T1.get_flags(T1.flags.integrated_prf)
    F2 = T2.get_flags(T2.flags.integrated_prf)
    assert F1 == F2
    I1 = I1.select(F1)
    I2 = I2.select(F2)
    assert flex.abs(I1 - I2) < 1e-6


def test_multi_lattice(dials_regression, tmpdir):
    expts = os.path.join(
        dials_regression, "integration_test_data", "multi_lattice", "experiments.json"
    )

    experiments = load.experiment_list(expts)
    for i, expt in enumerate(experiments):
        expt.identifier = str(100 + i)
    experiments.as_json(tmpdir / "modified_input.json")

    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "modified_input.json",
            os.path.join(
                dials_regression,
                "integration_test_data",
                "multi_lattice",
                "indexed.pickle",
            ),
            "prediction.padding=0",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("integrated.refl").check()
    assert tmpdir.join("integrated.expt").check()

    experiments = load.experiment_list(tmpdir / "integrated.expt")
    for i, expt in enumerate(experiments):
        assert expt.identifier == str(100 + i)

    table = flex.reflection_table.from_file(tmpdir / "integrated.refl")
    assert len(table) == 4962
    assert dict(table.experiment_identifiers()) == {0: "100", 1: "101"}
    # both should only have the imageset_id of zero as they share an imageset
    assert set(table["imageset_id"]) == {0}

    # Check output contains from two lattices
    exp_id = list(set(table["id"]))
    assert len(exp_id) == 2

    # Check both lattices have integrated reflections
    mask = table.get_flags(table.flags.integrated_prf)
    table = table.select(mask)
    exp_id = list(set(table["id"]))
    assert len(exp_id) == 2


def test_output_rubbish(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.index",
            dials_data("centroid_test_data", pathlib=True)
            / "imported_experiments.json",
            dials_data("centroid_test_data", pathlib=True) / "strong.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "indexed.expt").is_file()
    assert (tmp_path / "indexed.refl").is_file()

    # Call dials.integrate
    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "indexed.expt",
            "indexed.refl",
            "profile.fitting=False",
            "prediction.padding=0",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "integrated.refl").is_file()

    table = flex.reflection_table.from_file(tmp_path / "integrated.refl")
    assert "id" in table
    for row in table.rows():
        assert row["id"] == 0

    assert list(table.experiment_identifiers().keys()) == [0]
    assert list(table.experiment_identifiers().values())  # not empty


def test_integrate_with_kapton(dials_regression, tmpdir):
    pickle_name = "idx-20161021225550223_indexed.pickle"
    json_name = "idx-20161021225550223_refined_experiments.json"
    image_name = "20161021225550223.pickle"
    pickle_path = os.path.join(
        dials_regression, "integration_test_data", "stills_PSII", pickle_name
    )
    json_path = os.path.join(
        dials_regression, "integration_test_data", "stills_PSII", json_name
    )
    image_path = os.path.join(
        dials_regression, "integration_test_data", "stills_PSII", image_name
    )

    assert os.path.exists(pickle_path)
    assert os.path.exists(json_path)
    shutil.copy(pickle_path, tmpdir)
    shutil.copy(image_path, tmpdir)

    with open(tmpdir / json_name, "w") as w, open(json_path) as r:
        w.write(r.read() % tmpdir.strpath.replace("\\", "\\\\"))

    templ_phil = """
      output {
        experiments = 'idx-20161021225550223_integrated_experiments_%s.expt'
        reflections = 'idx-20161021225550223_integrated_%s.refl'
      }
      integration {
        lookup.mask = '%s'
        integrator = stills
        profile.fitting = False
        background.algorithm = simple
        debug {
          output = True
          separate_files = False
          split_experiments = False
        }
      }
      profile {
        gaussian_rs.min_spots.overall = 0
      }
      absorption_correction {
        apply = %s
        algorithm = fuller_kapton
        fuller_kapton {
          smart_sigmas = True
        }
      }
"""
    without_kapton_phil = templ_phil % (
        "nokapton",
        "nokapton",
        os.path.join(
            dials_regression, "integration_test_data", "stills_PSII", "mask.pickle"
        ).replace("\\", "\\\\"),
        "False",
    )
    with_kapton_phil = templ_phil % (
        "kapton",
        "kapton",
        os.path.join(
            dials_regression, "integration_test_data", "stills_PSII", "mask.pickle"
        ).replace("\\", "\\\\"),
        "True",
    )

    with open(tmpdir / "integrate_without_kapton.phil", "w") as f:
        f.write(without_kapton_phil)

    with open(tmpdir / "integrate_with_kapton.phil", "w") as f:
        f.write(with_kapton_phil)

    # Call dials.integrate with and without kapton correction
    for phil in "integrate_without_kapton.phil", "integrate_with_kapton.phil":
        result = procrunner.run(
            ["dials.integrate", "nproc=1", pickle_name, json_name, phil],
            working_directory=tmpdir,
        )
        assert not result.returncode and not result.stderr

    results = []
    for mode in "kapton", "nokapton":
        table = flex.reflection_table.from_file(
            tmpdir / f"idx-20161021225550223_integrated_{mode}.refl"
        )
        millers = table["miller_index"]
        test_indices = {"zero": (-5, 2, -6), "low": (-2, -20, 7), "high": (-1, -10, 4)}
        test_rows = {k: millers.first_index(v) for k, v in test_indices.items()}
        test_I_sigsqI = {
            k: (table[v]["intensity.sum.value"], table[v]["intensity.sum.variance"])
            for k, v in test_rows.items()
        }
        results.append(test_I_sigsqI)
    assert results[0]["zero"][0] == results[1]["zero"][0]
    assert results[0]["zero"][1] - results[1]["zero"][1] < 0.0001
    assert False not in [results[0]["low"][i] > results[1]["low"][i] for i in (0, 1)]
    assert False not in [results[0]["high"][i] > results[1]["high"][i] for i in (0, 1)]


def test_average_bbox_size():
    """Test behaviour of function for obtaining average bbox size."""
    reflections = flex.reflection_table()
    reflections["bbox"] = flex.int6(*(flex.int(10, i) for i in range(6)))
    assert _average_bbox_size(reflections) == (1, 1, 1)



from __future__ import annotations

import logging
import re

import libtbx
from libtbx.phil import parse

from dials.algorithms.refinement import DialsRefineConfigError
from dials.algorithms.refinement.constraints import phil_str as constr_phil_str
from dials.algorithms.refinement.parameterisation.autoreduce import (
    phil_str as autoreduce_phil_str,
)
from dials.algorithms.refinement.parameterisation.scan_varying_model_parameters import (
    phil_str as sv_phil_str,
)
from dials.algorithms.refinement.refinement_helpers import string_sel
from dials.algorithms.refinement.restraints.restraints_parameterisation import (
    uc_phil_str as uc_restraints_phil_str,
)

from .beam_parameters import BeamParameterisation
from .crystal_parameters import (
    CrystalOrientationParameterisation,
    CrystalUnitCellParameterisation,
)
from .detector_parameters import (
    DetectorParameterisationHierarchical,
    DetectorParameterisationMultiPanel,
    DetectorParameterisationSinglePanel,
)
from .goniometer_parameters import GoniometerParameterisation
from .prediction_parameters import (
    XYPhiPredictionParameterisation,
    XYPhiPredictionParameterisationSparse,
)
from .prediction_parameters_stills import (
    SphericalRelpStillsPredictionParameterisation,
    SphericalRelpStillsPredictionParameterisationSparse,
    StillsPredictionParameterisation,
    StillsPredictionParameterisationSparse,
)
from .scan_varying_beam_parameters import ScanVaryingBeamParameterisation
from .scan_varying_crystal_parameters import (
    ScanVaryingCrystalOrientationParameterisation,
    ScanVaryingCrystalUnitCellParameterisation,
)
from .scan_varying_detector_parameters import (
    ScanVaryingDetectorParameterisationSinglePanel,
)
from .scan_varying_goniometer_parameters import ScanVaryingGoniometerParameterisation
from .scan_varying_prediction_parameters import (
    ScanVaryingPredictionParameterisation,
    ScanVaryingPredictionParameterisationSparse,
)

format_data = {
    "autoreduce_phil": autoreduce_phil_str,
    "uc_restraints_phil": uc_restraints_phil_str,
    "constr_phil": constr_phil_str,
    "sv_phil": sv_phil_str,
}

logger = logging.getLogger(__name__)

phil_str = (
    """
    auto_reduction
      .help = "determine behaviour when there are too few reflections to"
              "reasonably produce a full parameterisation of the experiment list"
      .expert_level = 1
    {
      %(autoreduce_phil)s
    }

    scan_varying = False
      .help = "Allow models that are not forced to be static to vary during"
              "the scan, Auto will run one macrocycle with static then"
              "scan varying refinement for the crystal"
      .type = bool
      .short_caption = "Scan-varying refinement"

    compose_model_per = image *block
      .help = "For scan-varying parameterisations, compose a new model either"
              "every image or within blocks of a width specified in the"
              "reflections parameters. When this block width is larger than the"
              "image width the result is faster, with a trade-off in accuracy"
      .type = choice
      .expert_level = 1

    block_width = 1.0
      .help = "Width of a reflection 'block' (in degrees) determining how fine-"
              "grained the model used for scan-varying prediction during"
              "refinement is. Currently only has any effect if the crystal"
              "parameterisation is set to use compose_model_per=block"
      .type = float(value_min = 0.)
      .expert_level = 1

    set_scan_varying_errors = False
      .type = bool
      .help = "If scan-varying refinement is done, and if the estimated"
              "covariance of the model states have been calculated by the"
              "minimiser, choose whether to return this to the models or"
              "not. The default is not to, in order to keep the file size"
              "of the serialized model small. At the moment, this only has"
              "an effect for crystal unit cell (B matrix) errors."

    trim_scan_to_observations = True
      .type = bool
      .expert_level = 1
      .help = "For scan-varying refinement, trim scan objects to"
              "the range of observed reflections. This avoids"
              "failures in refinement for cases where the extremes of scans"
              "contain no data, such as when the crystal moves out of the beam."

    debug_centroid_analysis = False
      .help = "Set True to write out a file containing the reflections used"
              "for centroid analysis for automatic setting of the  scan-varying"
              "interval width. This can then be analysed with"
              "dev.dials.plot_centroid_analysis (requires dials_scratch repository)."
      .type = bool
      .expert_level = 2

    beam
      .help = "beam parameters"
    {
      fix = all *in_spindle_plane out_spindle_plane *wavelength
        .help = "Whether to fix beam parameters. By default,"
                "in_spindle_plane is selected, and one of the two"
                "parameters is fixed. If a goniometer is present"
                "this leads to the beam orientation being restricted"
                "to a direction in the initial spindle-beam plane."
                "Wavelength is also fixed by default, to allow refinement of"
                "the unit cell volume."
        .type = choice(multi=True)
        .short_caption = "Fix beam parameters"

      fix_list = None
        .type = strings
        .help = "Fix specified parameters by a list of 0-based indices or"
                "partial names to match"
        .expert_level = 1

      %(constr_phil)s

      force_static = True
        .type = bool
        .help = "Force a static parameterisation for the beam when doing"
                "scan-varying refinement"
        .expert_level = 1

      %(sv_phil)s
    }

    crystal
      .help = "crystal parameters"
    {
      fix = all cell orientation
        .help = "Fix crystal parameters"
        .type = choice
        .short_caption = "Fix crystal parameters"

      unit_cell
        .expert_level = 1
      {
        fix_list = None
          .type = strings
          .help = "Fix specified parameters by a list of 0-based indices or"
                  "partial names to match"
          .expert_level = 1

        %(uc_restraints_phil)s

        %(constr_phil)s

        force_static = False
          .type = bool
          .help = "Force a static parameterisation for the crystal unit cell"
                  "when doing scan-varying refinement"
          .expert_level = 1

        %(sv_phil)s
      }

      orientation
        .expert_level = 1
      {
        fix_list = None
          .type = strings
          .help = "Fix specified parameters by a list of 0-based indices or"
                  "partial names to match"
          .expert_level = 1

        %(constr_phil)s

        force_static = False
          .type = bool
          .help = "Force a static parameterisation for the crystal orientation"
                  "when doing scan-varying refinement"
          .expert_level = 1

        %(sv_phil)s
      }
    }

    detector
      .help = "detector parameters"
    {
      panels = *automatic single multiple hierarchical
        .help = "Select appropriate detector parameterisation. Both the"
                "single and multiple panel detector options treat the whole"
                "detector as a rigid body. The hierarchical parameterisation"
                "treats groups of panels as separate rigid bodies."
        .type = choice
        .expert_level = 1

      hierarchy_level = 0
        .help = "Level of the detector hierarchy (starting from the root at 0)"
                "at which to determine panel groups to parameterise independently"
        .type = int(value_min=0)
        .expert_level = 1

      fix = all position orientation distance
        .help = "Fix detector parameters. The translational parameters"
                "(position) may be set separately to the orientation."
        .type = choice
        .short_caption = "Fix detector parameters"

      fix_list = None
        .type = strings
        .help = "Fix specified parameters by a list of 0-based indices or"
                "partial names to match"
        .expert_level = 1

      %(constr_phil)s

      force_static = True
        .type = bool
        .help = "Force a static parameterisation for the detector"
                "when doing scan-varying refinement"
        .expert_level = 1

      %(sv_phil)s
    }

    goniometer
      .help = "goniometer setting matrix parameters"
    {
      fix = *all in_beam_plane out_beam_plane
        .help = "Whether to fix goniometer parameters. By default,"
                "fix all. Alternatively the setting matrix can be constrained"
                "to allow rotation only within the spindle-beam plane"
                "or to allow rotation only around an axis that lies in that"
                "plane. Set to None to refine the in two orthogonal directions."
        .type = choice(multi=True)
        .short_caption = "Fix goniometer parameters"

      fix_list = None
        .type = strings
        .help = "Fix specified parameters by a list of 0-based indices or"
                "partial names to match"
        .expert_level = 1

      %(constr_phil)s

      force_static = True
        .type = bool
        .help = "Force a static parameterisation for the goniometer when doing"
                "scan-varying refinement"
        .expert_level = 1

      %(sv_phil)s
    }

    sparse = Auto
      .help = "Calculate gradients using sparse data structures."
      .type = bool
      .expert_level = 1

    treat_single_image_as_still = False
      .help = "Set this to True to treat a single image scan with a non zero"
              "oscillation width as a still"
      .type = bool
      .expert_level = 1

    spherical_relp_model = False
      .help = "For stills refinement, set true to use the spherical relp model"
              "for prediction and gradients."
      .type = bool
      .expert_level = 1
"""
    % format_data
)
phil_scope = parse(phil_str)


# A helper function for parameter fixing
def _filter_parameter_names(parameterisation):
    # scan-varying suffixes like '_sample1' should be removed from
    # the list of parameter names so that it is num_free in length
    pattern = re.compile(r"_sample[0-9]+$")
    names = [
        pattern.sub("", e) for e in parameterisation.get_param_names(only_free=False)
    ]
    filtered_names = []
    for name in names:
        if name not in filtered_names:
            filtered_names.append(name)
    return filtered_names


# Helper function to perform centroid analysis
def _centroid_analysis(options, experiments, reflection_manager):

    analysis = None
    if not options.scan_varying:
        return analysis

    tst = [
        options.beam.smoother,
        options.crystal.orientation.smoother,
        options.crystal.unit_cell.smoother,
        options.detector.smoother,
        options.goniometer.smoother,
    ]
    tst = [
        (e.absolute_num_intervals is None and e.interval_width_degrees is libtbx.Auto)
        for e in tst
    ]
    if any(tst):
        logger.info(
            "Doing centroid analysis to "
            "automatically determine scan-varying interval widths"
        )
        ca = reflection_manager.get_centroid_analyser(
            debug=options.debug_centroid_analysis
        )
        analysis = ca()
    if analysis is None:
        return analysis

    # Use the results of centroid analysis to suggest suitable interval widths
    # for each experiment. This will be the smallest of the proposed intervals
    # for each of the residuals in x, y and phi, as long as this is not smaller
    # than either the outlier rejection block width, or 9.0 degrees.
    for i, a in enumerate(analysis):
        intervals = (a.get("x_interval"), a.get("y_interval"), a.get("phi_interval"))
        try:
            min_interval = min(_f for _f in intervals if _f is not None)
        except ValueError:
            # empty list - analysis was unable to suggest a suitable interval
            # width. Default to the safest case
            phi_min, phi_max = experiments[i].scan.get_oscillation_range(deg=True)
            a["interval_width"] = abs(phi_max - phi_min)
            logger.info(
                "Exp id {} suggested interval width could not be "
                "determined and will be reset to the scan width of "
                "{:.1f} degrees".format(i, a["interval_width"])
            )
            continue
        min_interval = max(min_interval, 9.0)
        block_size = a.get("block_size")
        if block_size is not None:
            min_interval = max(min_interval, block_size)
        a["interval_width"] = min_interval
        logger.info(
            "Exp id %s suggested interval width = %.1f degrees", i, min_interval
        )

    return analysis


# Helper function to choose the number of intervals for the smoother
def _set_n_intervals(smoother_params, analysis, scan, exp_ids):
    n_intervals = smoother_params.absolute_num_intervals
    if n_intervals is not None:
        return n_intervals

    deg_per_interval = smoother_params.interval_width_degrees
    if deg_per_interval is libtbx.Auto and analysis is not None:
        intervals = [analysis[i]["interval_width"] for i in exp_ids]
        deg_per_interval = min(intervals)
    if deg_per_interval is None:
        deg_per_interval = 36.0

    sequence_range_deg = scan.get_oscillation_range(deg=True)
    n_intervals = max(
        int(abs(sequence_range_deg[1] - sequence_range_deg[0]) / deg_per_interval), 1
    )
    return n_intervals


def _parameterise_beams(options, experiments, analysis):
    beam_params = []
    sv_beam = options.scan_varying and not options.beam.force_static
    for ibeam, beam in enumerate(experiments.beams()):
        # The Beam is parameterised with reference to a goniometer axis (or None).
        # Use the first (if any) Goniometers this Beam is associated with.
        exp_ids = experiments.indices(beam)
        assoc_models = [
            (experiments[i].goniometer, experiments[i].scan) for i in exp_ids
        ]
        goniometer, scan = assoc_models[0]

        if sv_beam:
            if not all((goniometer, scan)):
                raise DialsRefineConfigError(
                    "A scan-varying beam model cannot be created because "
                    "a scan or goniometer model is missing"
                )
            # If a beam is scan-varying, then it must always be found alongside
            # the same Scan and Goniometer in any Experiments in which it appears
            if not all(g is goniometer and s is scan for (g, s) in assoc_models):
                raise DialsRefineConfigError(
                    "A single scan-varying beam model cannot be refined "
                    "when associated with more than one scan or goniometer"
                )
            array_range = scan.get_array_range()
            n_intervals = _set_n_intervals(
                options.beam.smoother, analysis, scan, exp_ids
            )
            beam_param = ScanVaryingBeamParameterisation(
                beam,
                array_range,
                n_intervals,
                goniometer=goniometer,
                experiment_ids=exp_ids,
            )
        else:
            # Parameterise scan static beam, passing the goniometer
            beam_param = BeamParameterisation(beam, goniometer, experiment_ids=exp_ids)

        # Set the model identifier to name the parameterisation
        beam_param.model_identifier = f"Beam{ibeam + 1}"

        # get number of fixable units, either parameters or parameter sets in
        # the scan-varying case
        num_beam = getattr(beam_param, "num_sets", getattr(beam_param, "num_total"))()

        fix_list = []
        if options.beam.fix_list:
            fix_list.extend(options.beam.fix_list)

        if options.beam.fix:
            if "all" in options.beam.fix:
                beam_param.set_fixed([True] * num_beam)
            if "in_spindle_plane" in options.beam.fix:
                fix_list.append("Mu1")
            if "out_spindle_plane" in options.beam.fix:
                fix_list.append("Mu2")
            if "wavelength" in options.beam.fix:
                fix_list.append("nu")

        if fix_list:
            names = _filter_parameter_names(beam_param)
            assert len(names) == num_beam
            to_fix = string_sel(fix_list, names, beam_param.model_identifier)
            beam_param.set_fixed(to_fix)

        if beam_param.num_free() > 0:
            beam_params.append(beam_param)

    return beam_params


def _parameterise_crystals(options, experiments, analysis):
    xl_ori_params = []
    xl_uc_params = []
    sv_xl_ori = options.scan_varying and not options.crystal.orientation.force_static
    sv_xl_uc = options.scan_varying and not options.crystal.unit_cell.force_static
    for icrystal, crystal in enumerate(experiments.crystals()):
        # This crystal can only ever appear either in scans or in stills
        # (otherwise it requires a different crystal model)
        exp_ids = experiments.indices(crystal)
        assoc_models = [
            (experiments[i].goniometer, experiments[i].scan) for i in exp_ids
        ]
        goniometer, scan = assoc_models[0]
        if goniometer is None:
            # There should be no associated goniometer and scan models
            if any(g or s for (g, s) in assoc_models):
                raise DialsRefineConfigError(
                    "A crystal model appears in a mixture of scan and still "
                    "experiments, which is not supported"
                )

        if sv_xl_ori or sv_xl_uc:
            if not all((goniometer, scan)):
                raise DialsRefineConfigError(
                    "A scan-varying crystal model cannot be created because "
                    "a scan or goniometer model is missing"
                )
            # If a crystal is scan-varying, then it must always be found alongside
            # the same Scan and Goniometer in any Experiments in which it appears
            if not all(g is goniometer and s is scan for (g, s) in assoc_models):
                raise DialsRefineConfigError(
                    "A single scan-varying crystal model cannot be refined "
                    "when associated with more than one scan or goniometer"
                )
            array_range = scan.get_array_range()

        # orientation parameterisation
        if sv_xl_ori:
            n_intervals = _set_n_intervals(
                options.crystal.orientation.smoother, analysis, scan, exp_ids
            )
            xl_ori_param = ScanVaryingCrystalOrientationParameterisation(
                crystal, array_range, n_intervals, experiment_ids=exp_ids
            )
        else:  # force model to be static
            xl_ori_param = CrystalOrientationParameterisation(
                crystal, experiment_ids=exp_ids
            )

        # Set the model identifier to name the parameterisation
        xl_ori_param.model_identifier = f"Crystal{icrystal + 1}"

        # unit cell parameterisation
        if sv_xl_uc:
            n_intervals = _set_n_intervals(
                options.crystal.unit_cell.smoother, analysis, scan, exp_ids
            )
            set_errors = options.set_scan_varying_errors
            xl_uc_param = ScanVaryingCrystalUnitCellParameterisation(
                crystal,
                array_range,
                n_intervals,
                experiment_ids=exp_ids,
                set_state_uncertainties=set_errors,
            )
        else:  # force model to be static
            xl_uc_param = CrystalUnitCellParameterisation(
                crystal, experiment_ids=exp_ids
            )

        # Set the model identifier to name the parameterisation
        xl_uc_param.model_identifier = f"Crystal{icrystal + 1}"

        # get number of fixable units, either parameters or parameter sets in
        # the scan-varying case
        num_ori = getattr(
            xl_ori_param, "num_sets", getattr(xl_ori_param, "num_total")
        )()
        num_uc = getattr(xl_uc_param, "num_sets", getattr(xl_uc_param, "num_total"))()

        ori_fix_list = []
        if options.crystal.orientation.fix_list:
            ori_fix_list.extend(options.crystal.orientation.fix_list)

        cell_fix_list = []
        if options.crystal.unit_cell.fix_list:
            cell_fix_list.extend(options.crystal.unit_cell.fix_list)

        if options.crystal.fix:
            if options.crystal.fix == "all":
                xl_ori_param.set_fixed([True] * num_ori)
                xl_uc_param.set_fixed([True] * num_uc)
            elif options.crystal.fix == "cell":
                xl_uc_param.set_fixed([True] * num_uc)
            elif options.crystal.fix == "orientation":
                xl_ori_param.set_fixed([True] * num_ori)
            else:  # can only get here if refinement.phil is broken
                raise RuntimeError("crystal.fix value not recognised")

        if cell_fix_list:
            names = _filter_parameter_names(xl_uc_param)
            assert len(names) == num_uc
            to_fix = string_sel(cell_fix_list, names, xl_uc_param.model_identifier)
            xl_uc_param.set_fixed(to_fix)

        if ori_fix_list:
            names = _filter_parameter_names(xl_ori_param)
            assert len(names) == num_ori
            to_fix = string_sel(ori_fix_list, names, xl_ori_param.model_identifier)
            xl_ori_param.set_fixed(to_fix)

        if xl_ori_param.num_free() > 0:
            xl_ori_params.append(xl_ori_param)
        if xl_uc_param.num_free() > 0:
            xl_uc_params.append(xl_uc_param)

    return xl_ori_params, xl_uc_params


def _parameterise_detectors(options, experiments, analysis):
    det_params = []
    sv_det = options.scan_varying and not options.detector.force_static
    for idetector, detector in enumerate(experiments.detectors()):
        # keep associated gonio and scan in case we are scan-varying
        exp_ids = experiments.indices(detector)
        assoc_models = [
            (experiments[i].goniometer, experiments[i].scan) for i in exp_ids
        ]
        goniometer, scan = assoc_models[0]

        if sv_det:
            if not all((goniometer, scan)):
                raise DialsRefineConfigError(
                    "A scan-varying detector model cannot be created "
                    "because a scan or goniometer model is missing"
                )
            # If a detector is scan-varying, then it must always be found alongside
            # the same Scan and Goniometer in any Experiments in which it appears
            if not all(g is goniometer and s is scan for (g, s) in assoc_models):
                raise DialsRefineConfigError(
                    "A single scan-varying detector model cannot be "
                    "refined when associated with more than one scan or goniometer"
                )

            # Additional checks on whether a scan-varying parameterisation is allowed
            if options.detector.panels == "automatic" and len(detector) > 1:
                raise DialsRefineConfigError(
                    "Scan-varying multiple panel detectors are not "
                    "currently supported"
                )
            if options.detector.panels == "multiple":
                raise DialsRefineConfigError(
                    "Scan-varying multiple panel detectors are not "
                    "currently supported"
                )
            if options.detector.panels == "hierarchical":
                raise DialsRefineConfigError(
                    "Scan-varying hierarchical detectors are not " "currently supported"
                )

            array_range = scan.get_array_range()
            n_intervals = _set_n_intervals(
                options.detector.smoother, analysis, scan, exp_ids
            )
            det_param = ScanVaryingDetectorParameterisationSinglePanel(
                detector, array_range, n_intervals, experiment_ids=exp_ids
            )
        else:
            # Convert automatic into correct specific option
            if options.detector.panels == "automatic":
                if len(detector) > 1:
                    if hasattr(detector, "hierarchy"):
                        options.detector.panels = "hierarchical"
                    else:
                        options.detector.panels = "multiple"
                else:
                    options.detector.panels = "single"

            # Construct parameterisation based on panels choice
            if options.detector.panels == "single":
                if len(detector) > 1:
                    raise DialsRefineConfigError(
                        "A single panel parameterisation cannot be created "
                        "for a multiple panel detector"
                    )
                det_param = DetectorParameterisationSinglePanel(
                    detector, experiment_ids=exp_ids
                )
            elif options.detector.panels == "multiple":
                # Take first associated beam model
                beam = experiments[exp_ids[0]].beam
                det_param = DetectorParameterisationMultiPanel(
                    detector, beam, experiment_ids=exp_ids
                )
            elif options.detector.panels == "hierarchical":
                try:  # Use hierarchy in parameterisation if the detector has one
                    detector.hierarchy()
                    det_param = DetectorParameterisationHierarchical(
                        detector,
                        experiment_ids=exp_ids,
                        level=options.detector.hierarchy_level,
                    )
                except AttributeError:
                    raise DialsRefineConfigError(
                        "A hierarchical detector parameterisation cannot be "
                        "created for a detector without a hierarchy"
                    )

        # Set the model identifier to name the parameterisation
        det_param.model_identifier = f"Detector{idetector + 1}"

        # get number of fixable units, either parameters or parameter sets in
        # the scan-varying case
        num_det = getattr(det_param, "num_sets", getattr(det_param, "num_total"))()

        fix_list = []
        if options.detector.fix_list:
            fix_list.extend(options.detector.fix_list)

        if options.detector.fix:
            if options.detector.fix == "all":
                det_param.set_fixed([True] * num_det)
            elif options.detector.fix == "position":
                fix_list.extend(["Dist", "Shift1", "Shift2"])
            elif options.detector.fix == "orientation":
                fix_list.extend(["Tau"])
            elif options.detector.fix == "distance":
                fix_list.extend(["Dist", "Tau2", "Tau3"])
            else:  # can only get here if refinement.phil is broken
                raise RuntimeError("detector.fix value not recognised")

        if fix_list:
            names = _filter_parameter_names(det_param)
            assert len(names) == num_det
            to_fix = string_sel(fix_list, names, det_param.model_identifier)
            det_param.set_fixed(to_fix)

        if det_param.num_free() > 0:
            det_params.append(det_param)

    return det_params


def _parameterise_goniometers(options, experiments, analysis):
    gon_params = []
    sv_gon = options.scan_varying and not options.goniometer.force_static
    for igoniometer, goniometer in enumerate(experiments.goniometers()):
        if goniometer is None:
            continue
        # A Goniometer is parameterised with reference to the beam axis.
        # Use the first Beam this Goniometer is associated with.
        exp_ids = experiments.indices(goniometer)
        assoc_models = [(experiments[i].beam, experiments[i].scan) for i in exp_ids]
        beam, scan = assoc_models[0]

        if sv_gon:
            # If a goniometer is scan-varying, then it must always be found
            # alongside the same Scan in any Experiments in which it appears
            if not scan:
                raise DialsRefineConfigError(
                    "A scan-varying goniometer model cannot be created "
                    "because a scan model is missing"
                )
            if not all(s is scan for (g, s) in assoc_models):
                raise DialsRefineConfigError(
                    "A single scan-varying goniometer model cannot be "
                    "refined when associated with more than one scan"
                )
            array_range = scan.get_array_range()
            n_intervals = _set_n_intervals(
                options.goniometer.smoother, analysis, scan, exp_ids
            )
            gon_param = ScanVaryingGoniometerParameterisation(
                goniometer, array_range, n_intervals, beam=beam, experiment_ids=exp_ids
            )
        else:  # force model to be static
            gon_param = GoniometerParameterisation(
                goniometer, beam, experiment_ids=exp_ids
            )

        # Set the model identifier to name the parameterisation
        gon_param.model_identifier = f"Goniometer{igoniometer + 1}"

        # get number of fixable units, either parameters or parameter sets in
        # the scan-varying case
        num_gon = getattr(gon_param, "num_sets", getattr(gon_param, "num_total"))()

        fix_list = []
        if options.goniometer.fix_list:
            fix_list.extend(options.goniometer.fix_list)

        if options.goniometer.fix:
            if "all" in options.goniometer.fix:
                gon_param.set_fixed([True] * num_gon)
            if "in_beam_plane" in options.goniometer.fix:
                fix_list.append("Gamma1")
            if "out_beam_plane" in options.goniometer.fix:
                fix_list.append("Gamma2")

        if fix_list:
            names = _filter_parameter_names(gon_param)
            assert len(names) == num_gon
            to_fix = string_sel(fix_list, names, gon_param.model_identifier)
            gon_param.set_fixed(to_fix)

        if gon_param.num_free() > 0:
            gon_params.append(gon_param)

    return gon_params


def build_prediction_parameterisation(
    options, experiments, reflection_manager, do_stills=False
):
    """Given a set of parameters, create a parameterisation from a set of
    experimental models.

    Params:
        options: The input parameters
        experiments: An ExperimentList object
        reflection_manager: A ReflectionManager object
        do_stills (bool)

    Returns:
        A prediction equation parameterisation object
    """

    # If required, do full centroid analysis on the reflections (assumes
    # outlier-rejection has been done already) to determine suitable interval
    # widths for scan-varying refinement
    analysis = _centroid_analysis(options, experiments, reflection_manager)

    # Parameterise each unique model
    beam_params = _parameterise_beams(options, experiments, analysis)
    xl_ori_params, xl_uc_params = _parameterise_crystals(options, experiments, analysis)
    det_params = _parameterise_detectors(options, experiments, analysis)
    gon_params = _parameterise_goniometers(options, experiments, analysis)

    # Build the prediction equation parameterisation
    if do_stills:  # doing stills
        if options.sparse:
            if options.spherical_relp_model:
                PredParam = SphericalRelpStillsPredictionParameterisationSparse
            else:
                PredParam = StillsPredictionParameterisationSparse
        else:
            if options.spherical_relp_model:
                PredParam = SphericalRelpStillsPredictionParameterisation
            else:
                PredParam = StillsPredictionParameterisation
        pred_param = PredParam(
            experiments, det_params, beam_params, xl_ori_params, xl_uc_params
        )

    else:  # doing scans
        if options.scan_varying:
            if options.sparse:
                PredParam = ScanVaryingPredictionParameterisationSparse
            else:
                PredParam = ScanVaryingPredictionParameterisation
            pred_param = PredParam(
                experiments,
                det_params,
                beam_params,
                xl_ori_params,
                xl_uc_params,
                gon_params,
            )
            pred_param.set_scan_varying_errors = options.set_scan_varying_errors

        else:
            if options.sparse:
                PredParam = XYPhiPredictionParameterisationSparse
            else:
                PredParam = XYPhiPredictionParameterisation
            pred_param = PredParam(
                experiments,
                det_params,
                beam_params,
                xl_ori_params,
                xl_uc_params,
                gon_params,
            )

    return pred_param



from __future__ import annotations

from math import pi, sqrt

from cctbx.array_family import flex

from dials.algorithms.refinement.target import SparseGradientsMixin, Target

# constants
TWO_PI = 2.0 * pi


class LeastSquaresStillsResidualWithRmsdCutoff(Target):
    """An implementation of the target class providing a least squares residual
    in terms of detector impact position X, Y and minimum rotation to the Ewald
    sphere, DeltaPsi. Terminates refinement on achieved rmsd (or on intrisic
    convergence of the chosen minimiser)"""

    _grad_names = ["dX_dp", "dY_dp", "dDeltaPsi_dp"]
    rmsd_names = ["RMSD_X", "RMSD_Y", "RMSD_DeltaPsi"]
    rmsd_units = ["mm", "mm", "rad"]

    def __init__(
        self,
        experiments,
        predictor,
        reflection_manager,
        prediction_parameterisation,
        restraints_parameterisation,
        frac_binsize_cutoff=0.33333,
        absolute_cutoffs=None,
        gradient_calculation_blocksize=None,
    ):

        Target.__init__(
            self,
            experiments,
            predictor,
            reflection_manager,
            prediction_parameterisation,
            restraints_parameterisation,
            gradient_calculation_blocksize,
        )

        # Set up the RMSD achieved criterion. For simplicity, we take models from
        # the first Experiment only. If this is not appropriate for refinement over
        # all experiments then absolute cutoffs should be used instead.
        detector = experiments[0].detector
        if not absolute_cutoffs:
            pixel_sizes = [p.get_pixel_size() for p in detector]
            min_px_size_x = min(e[0] for e in pixel_sizes)
            min_px_size_y = min(e[1] for e in pixel_sizes)
            self._binsize_cutoffs = [
                min_px_size_x * frac_binsize_cutoff,
                min_px_size_y * frac_binsize_cutoff,
            ]
        else:
            self._binsize_cutoffs = absolute_cutoffs[:2]

        return

    def _predict_core(self, reflections):
        """perform prediction for the specified reflections"""

        # do prediction (updates reflection table in situ).
        self._reflection_predictor(reflections)

        x_obs, y_obs, _ = reflections["xyzobs.mm.value"].parts()
        x_calc, y_calc, _ = reflections["xyzcal.mm"].parts()

        # calculate residuals and assign columns
        reflections["x_resid"] = x_calc - x_obs
        reflections["x_resid2"] = reflections["x_resid"] ** 2
        reflections["y_resid"] = y_calc - y_obs
        reflections["y_resid2"] = reflections["y_resid"] ** 2
        reflections["delpsical2"] = reflections["delpsical.rad"] ** 2

        return reflections

    def predict_for_reflection_table(self, reflections, skip_derivatives=False):
        """perform prediction for all reflections in the supplied table"""

        # Just predict. The other preparation is irrelevant for stills, as is
        # skip_derivatives
        return self._predict_core(reflections)

    @staticmethod
    def _extract_residuals_and_weights(matches):

        # return residuals and weights as 1d flex.double vectors
        residuals = flex.double.concatenate(matches["x_resid"], matches["y_resid"])
        residuals.extend(matches["delpsical.rad"])

        weights, w_y, _ = matches["xyzobs.mm.weights"].parts()
        w_delpsi = matches["delpsical.weights"]
        weights.extend(w_y)
        weights.extend(w_delpsi)

        return residuals, weights

    @staticmethod
    def _extract_squared_residuals(matches):

        residuals2 = flex.double.concatenate(matches["x_resid2"], matches["y_resid2"])
        residuals2.extend(matches["delpsical2"])

        return residuals2

    def _rmsds_core(self, reflections):
        """calculate unweighted RMSDs"""

        resid_x = flex.sum(reflections["x_resid2"])
        resid_y = flex.sum(reflections["y_resid2"])
        resid_z = flex.sum(reflections["delpsical2"])
        n = len(reflections)
        rmsds = (sqrt(resid_x / n), sqrt(resid_y / n), sqrt(resid_z / n))
        return rmsds

    def achieved(self):
        """RMSD criterion for target achieved"""
        r = self._rmsds if self._rmsds else self.rmsds()

        # reset cached rmsds to avoid getting out of step
        self._rmsds = None

        # only use RMSD_X and RMSD_Y
        if r[0] < self._binsize_cutoffs[0] and r[1] < self._binsize_cutoffs[1]:
            return True
        return False


class LeastSquaresStillsResidualWithRmsdCutoffSparse(
    SparseGradientsMixin, LeastSquaresStillsResidualWithRmsdCutoff
):
    """A version of the LeastSquaresStillsResidualWithRmsdCutoff Target that
    uses a sparse matrix data structure for memory efficiency when there are a
    large number of Experiments"""

    pass



"""Test analytical expression for the partial derivatives of an angle
between two vectors with respect to each element of the vectors"""


from __future__ import annotations

import math
import random

from libtbx.test_utils import approx_equal
from scitbx import matrix
from scitbx.math import angle_derivative_wrt_vectors


class FDAngleDerivativeWrtVectorElts:
    """Given two vectors, u and v, calculate the derivative of the angle theta
    between them with respect to any of the elements u_1, u_2, u_3, v_1, v_2
    and v_3 using a finite difference approximation"""

    def __init__(self, u, v, delta=1.0e-7):
        self._vec_u = u
        self._vec_v = v
        self._u = u.length()
        self._v = v.length()
        self._delta = delta

    def dtheta_du_elt(self, i):
        """Return the derivative of theta with respect to the ith element of a"""

        half_delta_shift = [0.0, 0.0, 0.0]
        half_delta_shift[i] = self._delta / 2.0
        half_delta_shift = matrix.col(half_delta_shift)

        u_fwd = self._vec_u + half_delta_shift
        u_rev = self._vec_u - half_delta_shift

        theta_fwd = math.acos(u_fwd.dot(self._vec_v) / (u_fwd.length() * self._v))
        theta_rev = math.acos(u_rev.dot(self._vec_v) / (u_rev.length() * self._v))

        return (theta_fwd - theta_rev) / self._delta

    def dtheta_dv_elt(self, i):
        """Return the derivative of theta with respect to the ith element of b"""

        half_delta_shift = [0.0, 0.0, 0.0]
        half_delta_shift[i] = self._delta / 2.0
        half_delta_shift = matrix.col(half_delta_shift)

        v_fwd = self._vec_v + half_delta_shift
        v_rev = self._vec_v - half_delta_shift

        theta_fwd = math.acos(self._vec_u.dot(v_fwd) / (v_fwd.length() * self._u))
        theta_rev = math.acos(self._vec_u.dot(v_rev) / (v_rev.length() * self._u))

        return (theta_fwd - theta_rev) / self._delta

    def dtheta_du_1(self):
        return self.dtheta_du_elt(0)

    def dtheta_du_2(self):
        return self.dtheta_du_elt(1)

    def dtheta_du_3(self):
        return self.dtheta_du_elt(2)

    def dtheta_dv_1(self):
        return self.dtheta_dv_elt(0)

    def dtheta_dv_2(self):
        return self.dtheta_dv_elt(1)

    def dtheta_dv_3(self):
        return self.dtheta_dv_elt(2)


def _test():
    # Two random vectors
    vec_a = matrix.col(
        (random.uniform(-20, 20), random.uniform(-20, 20), random.uniform(-20, 20))
    )
    vec_b = matrix.col(
        (random.uniform(-20, 20), random.uniform(-20, 20), random.uniform(-20, 20))
    )

    # The test may fail if the angle between the vectors is small or close to pi
    # (see gradient of acos(x) for x close to 1 or -1) but we are not interested
    # in such cases anyway. Skip test if the acute angle is less than 1 degree
    if vec_a.accute_angle(vec_b, deg=True) < 1:
        return False

    # Analytical
    def dangle(u, v):
        return [matrix.col(e) for e in angle_derivative_wrt_vectors(u, v)]

    dgamma_da, dgamma_db = dangle(vec_a, vec_b)

    # FD
    dgFD = FDAngleDerivativeWrtVectorElts(vec_a, vec_b)
    assert approx_equal(
        dgamma_da,
        matrix.col([dgFD.dtheta_du_1(), dgFD.dtheta_du_2(), dgFD.dtheta_du_3()]),
    )
    assert approx_equal(
        dgamma_db,
        matrix.col([dgFD.dtheta_dv_1(), dgFD.dtheta_dv_2(), dgFD.dtheta_dv_3()]),
    )

    return True


def test_should_succeed_in_95_percent_of_cases():
    ntests = 1000
    results = [_test() for i in range(ntests)]
    nsuccess = results.count(True)
    assert nsuccess > 0.95 * ntests



from __future__ import annotations

import os
import sys

try:
    import stat
except ImportError:
    stat = None


def nag():
    """
    Check if pre-commits should be installed for this repository.
    If they are not and should be then annoy the developer.
    To be called in libtbx_refresh.py
    """
    if os.name == "nt" or not stat:  # unsupported
        return
    # Determine the name of the calling module, and thus the internal module name
    # of the libtbx_refresh file. Use exception trick to pick up the current frame.
    try:
        raise Exception()
    except Exception:
        frame = sys.exc_info()[2].tb_frame.f_back
    # Extract the caller name
    caller = frame.f_globals["__name__"]
    if caller == "__main__":
        # well that is not very informative, is it.
        caller = os.path.abspath(
            frame.f_code.co_filename
        )  # Get the full path of the libtbx_refresh.py file.
        refresh_file, _ = os.path.splitext(caller)
        if not refresh_file.endswith("libtbx_refresh"):
            raise RuntimeError(
                "pre-commit nagging can only be done from within libtbx_refresh.py"
            )
        # the name of the parent directory of libtbx_refresh.py is the caller name
        caller = os.path.basename(os.path.dirname(refresh_file))
    else:
        if not caller.endswith(".libtbx_refresh"):
            raise RuntimeError(
                "pre-commit nagging can only be done from within libtbx_refresh.py"
            )
        caller = caller[:-15]

    try:
        import libtbx.load_env
    except Exception as e:
        print("error on importing libtbx environment for pre-commit nagging:", e)
        return
    try:
        path = libtbx.env.dist_path(caller)
    except Exception as e:
        print(f"error on obtaining module path for {caller} for pre-commit nagging:", e)
        return

    if not os.path.isdir(os.path.join(path, ".git")):
        return  # not a developer installation

    precommit_python = abs(libtbx.env.build_path / "precommitbx" / "bin" / "python3")
    hookfile = os.path.join(path, ".git", "hooks", "pre-commit")
    if os.path.isfile(hookfile) and os.access(hookfile, os.X_OK):
        with open(hookfile) as fh:
            precommit = fh.read()
        if "precommitbx" in precommit and os.path.exists(precommit_python):
            return  # legacy libtbx.precommit hook is fine
        if "precommitbx for conda" in precommit:
            return  # libtbx.precommit hook with conda environment is fine
        if "generated by pre-commit" in precommit and "libtbx" not in precommit:
            return  # genuine pre-commit hook is also fine

    try:
        with open(hookfile, "w") as fh:
            fh.write(
                """#!/bin/bash
echo
echo Please install the DIALS pre-commit hooks before committing into the DIALS
echo repository. These hooks run simple static code analysis to catch common
echo coding mistakes early and ensure a common code style.
echo
echo The command you need to run is:
echo "  libtbx.precommit install"
echo

if [ -z "$DIALS_WITHOUT_PRECOMMITS" ]; then
  echo If you want to continue without installing pre-commits then you can override
  echo this check by setting the environment variable DIALS_WITHOUT_PRECOMMITS
fi
echo You can find more information about contributing to DIALS at:
echo https://github.com/dials/dials/blob/main/CONTRIBUTING.md
echo
if [ -z "$DIALS_WITHOUT_PRECOMMITS" ]; then
  exit 1
fi
"""
            )
            mode = os.fstat(fh.fileno()).st_mode
            mode |= stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
            os.fchmod(fh.fileno(), stat.S_IMODE(mode))
    except Exception as e:
        print("Could not generate pre-commit stub:", e)



from __future__ import annotations

from math import sqrt

import numpy as np
import pytest
from numpy.linalg import norm

from dxtbx import flumpy
from scitbx import matrix

from dials.algorithms.profile_model.ellipsoid import chisq_quantile
from dials.algorithms.profile_model.ellipsoid.model import (
    Simple1ProfileModel,
    Simple6ProfileModel,
    compute_change_of_basis_operation,
)
from dials.algorithms.profile_model.ellipsoid.parameterisation import (
    ModelState,
    Simple1MosaicityParameterisation,
    Simple6MosaicityParameterisation,
)
from dials.algorithms.spot_prediction import IndexGenerator
from dials.array_family import flex
from dials.constants import FULL_PARTIALITY


@pytest.fixture
def simple1_profile_model():

    params = flex.double([4e-4])

    model = Simple1ProfileModel(params)
    return model


@pytest.fixture
def simple6_profile_model():
    params = flex.double([4e-4, 1e-4, 5e-4, 2e-4, 3e-4, 6e-4])
    model = Simple6ProfileModel(params)
    return model


@pytest.fixture
def simple1_model_state(test_experiment):

    state = ModelState(test_experiment, Simple1MosaicityParameterisation())

    return state


@pytest.fixture
def simple6_model_state(test_experiment):

    state = ModelState(test_experiment, Simple6MosaicityParameterisation())

    return state


def check_simple1_sigma(sigma, params):
    b1 = params[0]

    assert sigma[0, 0] == pytest.approx(b1**2)
    assert sigma[0, 1] == 0
    assert sigma[0, 2] == 0
    assert sigma[1, 0] == 0
    assert sigma[1, 1] == pytest.approx(b1**2)
    assert sigma[1, 2] == 0
    assert sigma[2, 0] == 0
    assert sigma[2, 1] == 0
    assert sigma[2, 2] == pytest.approx(b1**2)


def check_simple6_sigma(sigma, params):

    b1, b2, b3, b4, b5, b6 = params
    L = np.array([[b1, 0, 0], [b2, b3, 0], [b4, b5, b6]])
    M = np.matmul(L, L.T)
    for i in range(3):
        for j in range(3):
            assert sigma[i, j] == pytest.approx(M[i, j])


def test_Simple1ProfileModel_sigma(simple1_profile_model):
    sigma = simple1_profile_model.sigma()
    check_simple1_sigma(sigma, [4e-4])


def test_Simple1ProfileModel_update_model(simple1_profile_model, simple1_model_state):
    params = [5e-4]
    simple1_model_state.M_params = params
    simple1_profile_model.update_model(simple1_model_state)
    sigma = simple1_profile_model.sigma()
    check_simple1_sigma(sigma, params)


def test_Simple1ProfileModel_predict_reflections(
    simple1_profile_model,
    test_experiment,
):

    # Create the index generator
    index_generator = IndexGenerator(
        test_experiment.crystal.get_unit_cell(),
        test_experiment.crystal.get_space_group().type(),
        2.0,
    )

    # Get an array of miller indices
    miller_indices = index_generator.to_array()
    reflections = simple1_profile_model.predict_reflections(
        [test_experiment], miller_indices, probability=FULL_PARTIALITY
    )

    s0 = matrix.col(test_experiment.beam.get_s0())
    quantile = chisq_quantile(3, FULL_PARTIALITY)
    sigma_inv = matrix.sqr(flumpy.from_numpy(simple1_profile_model.sigma())).inverse()

    for s2 in reflections["s2"]:
        s2_ = matrix.col(s2)
        x = s2_.normalize() * s0.length() - s2_
        d = (x.transpose() * sigma_inv * x)[0]
        assert d < quantile


def test_Simple1ProfileModel_compute_bbox(simple1_profile_model, test_experiment):

    experiments = [test_experiment]

    # Create the index generator
    index_generator = IndexGenerator(
        experiments[0].crystal.get_unit_cell(),
        experiments[0].crystal.get_space_group().type(),
        2.0,
    )

    # Get an array of miller indices
    miller_indices = index_generator.to_array()
    reflections = simple1_profile_model.predict_reflections(
        experiments, miller_indices, probability=FULL_PARTIALITY
    )

    simple1_profile_model.compute_bbox(experiments, reflections)


def test_Simple1ProfileModel_compute_mask(simple1_profile_model, test_experiment):
    experiments = [test_experiment]

    # Create the index generator
    index_generator = IndexGenerator(
        experiments[0].crystal.get_unit_cell(),
        experiments[0].crystal.get_space_group().type(),
        2.0,
    )

    # Get an array of miller indices
    miller_indices = index_generator.to_array()
    reflections = simple1_profile_model.predict_reflections(
        experiments, miller_indices, probability=FULL_PARTIALITY
    )

    simple1_profile_model.compute_bbox(experiments, reflections)

    reflections["shoebox"] = flex.shoebox(
        reflections["panel"], reflections["bbox"], allocate=True
    )

    simple1_profile_model.compute_mask(experiments, reflections)


def test_Simple1ProfileModel_sigma_for_reflection(simple1_profile_model):
    sigma = simple1_profile_model.sigma()
    check_simple1_sigma(sigma, [4e-4])


def test_Simple1ProfileModel_parameterisation(simple1_profile_model):
    assert isinstance(
        simple1_profile_model.parameterisation(), Simple1MosaicityParameterisation
    )


def test_Simple6ProfileModel_sigma(simple6_profile_model):
    sigma = simple6_profile_model.sigma()
    check_simple6_sigma(sigma, [4e-4, 1e-4, 5e-4, 2e-4, 3e-4, 6e-4])


def test_Simple6ProfileModel_update_model(simple6_profile_model, simple6_model_state):
    params = [5e-4, 2e-4, 6e-4, 3e-4, 4e-4, 7e-4]
    simple6_model_state.M_params = params
    simple6_profile_model.update_model(simple6_model_state)
    sigma = simple6_profile_model.sigma()
    check_simple6_sigma(sigma, params)


def test_Simple6ProfileModel_predict_reflections(
    simple6_profile_model, test_experiment
):
    experiments = [test_experiment]

    # Create the index generator
    index_generator = IndexGenerator(
        experiments[0].crystal.get_unit_cell(),
        experiments[0].crystal.get_space_group().type(),
        2.0,
    )

    # Get an array of miller indices
    miller_indices = index_generator.to_array()
    reflections = simple6_profile_model.predict_reflections(
        experiments, miller_indices, probability=FULL_PARTIALITY
    )

    s2 = reflections["s2"]
    s0 = matrix.col(experiments[0].beam.get_s0())
    quantile = chisq_quantile(3, FULL_PARTIALITY)
    sigma_inv = matrix.sqr(flumpy.from_numpy(simple6_profile_model.sigma())).inverse()

    for s2 in map(matrix.col, reflections["s2"]):
        x = s2.normalize() * s0.length() - s2
        d = (x.transpose() * sigma_inv * x)[0]
        assert d < quantile


def test_Simple6ProfileModel_compute_bbox(simple6_profile_model, test_experiment):
    experiments = [test_experiment]

    # Create the index generator
    index_generator = IndexGenerator(
        experiments[0].crystal.get_unit_cell(),
        experiments[0].crystal.get_space_group().type(),
        2.0,
    )

    # Get an array of miller indices
    miller_indices = index_generator.to_array()
    reflections = simple6_profile_model.predict_reflections(
        experiments, miller_indices, probability=FULL_PARTIALITY
    )

    s2 = reflections["s2"]
    s0 = matrix.col(experiments[0].beam.get_s0())
    quantile = chisq_quantile(3, FULL_PARTIALITY)
    sigma_inv = matrix.sqr(flumpy.from_numpy(simple6_profile_model.sigma())).inverse()

    for s2 in map(matrix.col, reflections["s2"]):
        x = s2.normalize() * s0.length() - s2
        d = (x.transpose() * sigma_inv * x)[0]
        assert d < quantile

    simple6_profile_model.compute_bbox(experiments, reflections)


def test_Simple6ProfileModel_compute_mask(simple6_profile_model, test_experiment):
    experiments = [test_experiment]

    # Create the index generator
    index_generator = IndexGenerator(
        experiments[0].crystal.get_unit_cell(),
        experiments[0].crystal.get_space_group().type(),
        2.0,
    )

    # Get an array of miller indices
    miller_indices = index_generator.to_array()
    reflections = simple6_profile_model.predict_reflections(
        experiments, miller_indices, probability=FULL_PARTIALITY
    )

    s2 = reflections["s2"]
    s0 = matrix.col(experiments[0].beam.get_s0())
    quantile = chisq_quantile(3, FULL_PARTIALITY)
    sigma_inv = matrix.sqr(flumpy.from_numpy(simple6_profile_model.sigma())).inverse()

    for s2 in map(matrix.col, reflections["s2"]):
        x = s2.normalize() * s0.length() - s2
        d = (x.transpose() * sigma_inv * x)[0]
        assert d < quantile

    simple6_profile_model.compute_bbox(experiments, reflections)

    reflections["shoebox"] = flex.shoebox(
        reflections["panel"], reflections["bbox"], allocate=True
    )

    simple6_profile_model.compute_mask(experiments, reflections)


def test_Simple6ProfileModel_sigma_for_reflection(simple6_profile_model):
    sigma = simple6_profile_model.sigma()
    check_simple6_sigma(sigma, [4e-4, 1e-4, 5e-4, 2e-4, 3e-4, 6e-4])


def test_Simple6ProfileModel_parameterisation(simple6_profile_model):
    assert isinstance(
        simple6_profile_model.parameterisation(), Simple6MosaicityParameterisation
    )


def test_compute_change_of_basis_operation():

    r = np.array((0, 0.5, -(1 - sqrt(0.75))))
    s0 = np.array((0, 0, 1))
    s2 = s0 + r

    R = compute_change_of_basis_operation(s0, s2)

    z = np.matmul(R, s2)

    assert norm(z) == pytest.approx(1)
